@@ -274,6 +274,7 @@ cmd_info(int argc, char *argv[])
 	int ch, npackfiles, npackedobj, nobj;
 	off_t packsize, loose_size;
 	char scaled[FMT_SCALED_STRSIZE];
+	int *pack_fds = NULL;
 
 	while ((ch = getopt(argc, argv, "r:")) != -1) {
 		switch (ch) {
@@ -294,7 +295,7 @@ cmd_info(int argc, char *argv[])
 	argv += optind;
 
 #ifndef PROFILE
-	if (pledge("stdio rpath wpath flock proc exec sendfd unveil",
+	if (pledge("stdio rpath wpath cpath flock proc exec sendfd unveil",
 	    NULL) == -1)
 		err(1, "pledge");
 #endif
@@ -303,10 +304,18 @@ cmd_info(int argc, char *argv[])
 		if (error)
 			goto done;
 	}
-	error = got_repo_open(&repo, repo_path, NULL);
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, repo_path, NULL, pack_fds);
 	if (error)
 		goto done;
-
+#ifndef PROFILE
+	/* Remove "cpath" promise. */
+	if (pledge("stdio rpath wpath flock proc exec sendfd unveil",
+	    NULL) == -1)
+		err(1, "pledge");
+#endif
 	error = apply_unveil(got_repo_get_path_git_dir(repo), 1);
 	if (error)
 		goto done;
@@ -365,6 +374,13 @@ cmd_info(int argc, char *argv[])
 done:
 	if (repo)
 		got_repo_close(repo);
+	if (pack_fds) {
+		const struct got_error *pack_err =
+		    got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
+
 	free(repo_path);
 	return error;
 }
@@ -623,6 +639,7 @@ cmd_pack(int argc, char *argv[])
 	struct got_reflist_head exclude_refs;
 	struct got_reflist_head include_refs;
 	struct got_reflist_entry *re, *new;
+	int *pack_fds = NULL;
 
 	TAILQ_INIT(&exclude_args);
 	TAILQ_INIT(&exclude_refs);
@@ -669,7 +686,10 @@ cmd_pack(int argc, char *argv[])
 		if (error)
 			goto done;
 	}
-	error = got_repo_open(&repo, repo_path, NULL);
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, repo_path, NULL, pack_fds);
 	if (error)
 		goto done;
 
@@ -741,6 +761,12 @@ cmd_pack(int argc, char *argv[])
 done:
 	if (repo)
 		got_repo_close(repo);
+	if (pack_fds) {
+		const struct got_error *pack_err =
+		    got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
 	got_pathlist_free(&exclude_args);
 	got_ref_list_free(&exclude_refs);
 	got_ref_list_free(&include_refs);
@@ -769,6 +795,7 @@ cmd_indexpack(int argc, char *argv[])
 	char *id_str = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
+	int *pack_fds = NULL;
 
 	while ((ch = getopt(argc, argv, "")) != -1) {
 		switch (ch) {
@@ -794,7 +821,10 @@ cmd_indexpack(int argc, char *argv[])
 		err(1, "pledge");
 #endif
 
-	error = got_repo_open(&repo, packfile_path, NULL);
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, packfile_path, NULL, pack_fds);
 	if (error)
 		goto done;
 
@@ -824,6 +854,12 @@ cmd_indexpack(int argc, char *argv[])
 done:
 	if (repo)
 		got_repo_close(repo);
+	if (pack_fds) {
+		const struct got_error *pack_err =
+		    got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
 	free(id_str);
 	free(pack_hash);
 	return error;
@@ -856,7 +892,7 @@ list_pack_cb(void *arg, struct got_object_id *id, int type, off_t offset,
 	char *id_str, *delta_str = NULL, *base_id_str = NULL;
 	const char *type_str;
 
-	err = got_object_id_str(&id_str, id);	
+	err = got_object_id_str(&id_str, id);
 	if (err)
 		return err;
 
@@ -937,6 +973,7 @@ cmd_listpack(int argc, char *argv[])
 	struct gotadmin_list_pack_cb_args lpa;
 	FILE *packfile = NULL;
 	int show_stats = 0, human_readable = 0;
+	int *pack_fds = NULL;
 
 	while ((ch = getopt(argc, argv, "hs")) != -1) {
 		switch (ch) {
@@ -962,14 +999,22 @@ cmd_listpack(int argc, char *argv[])
 		return got_error_from_errno2("realpath", argv[0]);
 
 #ifndef PROFILE
+	if (pledge("stdio rpath wpath cpath flock proc exec sendfd unveil",
+	    NULL) == -1)
+		err(1, "pledge");
+#endif
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, packfile_path, NULL, pack_fds);
+	if (error)
+		goto done;
+#ifndef PROFILE
+	/* Remove "cpath" promise. */
 	if (pledge("stdio rpath wpath flock proc exec sendfd unveil",
 	    NULL) == -1)
 		err(1, "pledge");
 #endif
-	error = got_repo_open(&repo, packfile_path, NULL);
-	if (error)
-		goto done;
-
 	error = apply_unveil(got_repo_get_path_git_dir(repo), 1);
 	if (error)
 		goto done;
@@ -999,6 +1044,12 @@ cmd_listpack(int argc, char *argv[])
 done:
 	if (repo)
 		got_repo_close(repo);
+	if (pack_fds) {
+		const struct got_error *pack_err =
+		    got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
 	free(id_str);
 	free(pack_hash);
 	free(packfile_path);
@@ -1109,6 +1160,7 @@ cmd_cleanup(int argc, char *argv[])
 	char scaled_diff[FMT_SCALED_STRSIZE];
 	char **extensions;
 	int nextensions, i;
+	int *pack_fds = NULL;
 
 	while ((ch = getopt(argc, argv, "apr:nq")) != -1) {
 		switch (ch) {
@@ -1150,7 +1202,10 @@ cmd_cleanup(int argc, char *argv[])
 		if (error)
 			goto done;
 	}
-	error = got_repo_open(&repo, repo_path, NULL);
+	error = got_repo_pack_fds_open(&pack_fds);
+	if (error != NULL)
+		goto done;
+	error = got_repo_open(&repo, repo_path, NULL, pack_fds);
 	if (error)
 		goto done;
 
@@ -1215,6 +1270,12 @@ cmd_cleanup(int argc, char *argv[])
 done:
 	if (repo)
 		got_repo_close(repo);
+	if (pack_fds) {
+		const struct got_error *pack_err =
+		    got_repo_pack_fds_close(pack_fds);
+		if (error == NULL)
+			error = pack_err;
+	}
 	free(repo_path);
 	return error;
 }
