@@ -16,8 +16,15 @@
 
 /*
  * Compute the differences between two blobs and write unified diff text
- * to the provided output FILE. Two const char * diff header labels may
- * be provided which will be used to identify each blob in the diff output.
+ * to the provided output file. Two open temporary files must be provided
+ * for internal use; these files can be obtained from got_opentemp() and
+ * must be closed by the caller.
+ * If one of the blobs being diffed does not exist, all corresponding
+ * blob object and temporary file arguments should be set to NULL.
+ * Two const char * diff header labels may be provided which will be used
+ * to identify each blob in the diff output.
+ * The set of arguments relating to either blob may be NULL to indicate
+ * that no content is present on its respective side of the diff.
  * If a label is NULL, use the blob's SHA1 checksum instead.
  * The number of context lines to show in the diff must be specified as well.
  * Whitespace differences may optionally be ignored.
@@ -25,19 +32,20 @@
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_blob(off_t **, size_t *,
-    struct got_blob_object *, struct got_blob_object *,
+    struct got_blob_object *, struct got_blob_object *, FILE *, FILE *,
     const char *, const char *, int, int, int, FILE *);
 
 /*
  * Compute the differences between a blob and a file and write unified diff
- * text to the provided output file. The file's size must be provided, as
- * well as a const char * diff header label which identifies the file.
+ * text to the provided output file. The blob object, its content, and its
+ * size must be provided.The file's size must be provided, as well as a
+ * const char * diff header label which identifies the file.
  * An optional const char * diff header label for the blob may be provided, too.
  * The number of context lines to show in the diff must be specified as well.
  * Whitespace differences may optionally be ignored.
  */
-const struct got_error *got_diff_blob_file(struct got_blob_object *,
-    const char *, FILE *, size_t, const char *, int, int, int, FILE *);
+const struct got_error *got_diff_blob_file(struct got_blob_object *, FILE *,
+    off_t, const char *, FILE *, size_t, const char *, int, int, int, FILE *);
 
 /*
  * A callback function invoked to handle the differences between two blobs
@@ -45,13 +53,15 @@ const struct got_error *got_diff_blob_file(struct got_blob_object *,
  * their respective IDs, and two corresponding paths within the diffed trees.
  * The first blob contains content from the old side of the diff, and
  * the second blob contains content on the new side of the diff.
+ * Two open temporary files must be provided for internal use; these files
+ * can be obtained from got_opentemp() and must be closed by the caller.
  * The set of arguments relating to either blob may be NULL to indicate
  * that no content is present on its respective side of the diff.
  * File modes from relevant tree objects which contain the blobs may
  * also be passed. These will be zero if not available.
  */
 typedef const struct got_error *(*got_diff_blob_cb)(void *,
-    struct got_blob_object *, struct got_blob_object *,
+    struct got_blob_object *, struct got_blob_object *, FILE *, FILE *,
     struct got_object_id *, struct got_object_id *,
     const char *, const char *, mode_t, mode_t, struct got_repository *);
 
@@ -82,7 +92,7 @@ struct got_diff_blob_output_unidiff_arg {
 	off_t *line_offsets;	/* Dispose of with free(3) when done. */
 };
 const struct got_error *got_diff_blob_output_unidiff(void *,
-    struct got_blob_object *, struct got_blob_object *,
+    struct got_blob_object *, struct got_blob_object *, FILE *, FILE *,
     struct got_object_id *, struct got_object_id *,
     const char *, const char *, mode_t, mode_t, struct got_repository *);
 
@@ -92,9 +102,14 @@ const struct got_error *got_diff_blob_output_unidiff(void *,
  * Diffing of blob content can be suppressed by passing zero for the
  * 'diff_content' parameter. The callback will then only receive blob
  * object IDs and diff labels, but NULL pointers instead of blob objects.
+ * If 'diff_content' is set, two open temporary files must be provided
+ * for internal use; these files can be obtained from got_opentemp()
+ * and must be closed by the caller. Otherwise the files can be NULL.
+ * The set of arguments relating to either tree may be NULL to indicate
+ * that no content is present on its respective side of the diff.
  */
 const struct got_error *got_diff_tree(struct got_tree_object *,
-    struct got_tree_object *, const char *, const char *,
+    struct got_tree_object *, FILE *, FILE *, const char *, const char *,
     struct got_repository *, got_diff_blob_cb cb, void *cb_arg, int);
 
 /*
@@ -114,7 +129,7 @@ struct got_diff_changed_path {
 	int status;
 };
 const struct got_error *got_diff_tree_collect_changed_paths(void *,
-    struct got_blob_object *, struct got_blob_object *,
+    struct got_blob_object *, struct got_blob_object *, FILE *, FILE *,
     struct got_object_id *, struct got_object_id *,
     const char *, const char *, mode_t, mode_t, struct got_repository *);
 
@@ -122,13 +137,17 @@ const struct got_error *got_diff_tree_collect_changed_paths(void *,
  * Diff two objects, assuming both objects are blobs. Two const char * diff
  * header labels may be provided which will be used to identify each blob in
  * the diff output. If a label is NULL, use the blob's SHA1 checksum instead.
+ * Two open temporary files must be provided for internal use; these files
+ * can be obtained from got_opentemp() and must be closed by the caller.
+ * The set of arguments relating to either blob may be NULL to indicate
+ * that no content is present on its respective side of the diff.
  * The number of context lines to show in the diff must be specified as well.
  * Write unified diff text to the provided output FILE.
  * If not NULL, the two initial output arguments will be populated with an
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_objects_as_blobs(off_t **, size_t *,
-    struct got_object_id *, struct got_object_id *,
+    FILE *, FILE *, struct got_object_id *, struct got_object_id *,
     const char *, const char *, int, int, int,
     struct got_repository *, FILE *);
 
@@ -137,23 +156,32 @@ const struct got_error *got_diff_objects_as_blobs(off_t **, size_t *,
  * header labels may be provided which will be used to identify each blob in
  * the trees. If a label is NULL, use the blob's SHA1 checksum instead.
  * The number of context lines to show in diffs must be specified.
+ * Two open temporary files must be provided for internal use; these files
+ * can be obtained from got_opentemp() and must be closed by the caller.
+ * The set of arguments relating to either tree may be NULL to indicate
+ * that no content is present on its respective side of the diff.
  * Write unified diff text to the provided output FILE.
  * If not NULL, the two initial output arguments will be populated with an
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_objects_as_trees(off_t **, size_t *,
-    struct got_object_id *, struct got_object_id *, struct got_pathlist_head *,
-    char *, char *, int, int, int, struct got_repository *, FILE *);
+    FILE *, FILE *, struct got_object_id *, struct got_object_id *,
+    struct got_pathlist_head *, char *, char *, int, int, int,
+    struct got_repository *, FILE *);
 
 /*
  * Diff two objects, assuming both objects are commits.
  * The number of context lines to show in diffs must be specified.
+ * Two open temporary files must be provided for internal use; these files
+ * can be obtained from got_opentemp() and must be closed by the caller.
+ * The set of arguments relating to either commit may be NULL to indicate
+ * that no content is present on its respective side of the diff.
  * Write unified diff text to the provided output FILE.
  * If not NULL, the two initial output arguments will be populated with an
  * array of line offsets for, and the number of lines in, the unidiff text.
  */
 const struct got_error *got_diff_objects_as_commits(off_t **, size_t *,
-    struct got_object_id *, struct got_object_id *, struct got_pathlist_head *,
-    int, int, int, struct got_repository *, FILE *);
+    FILE *, FILE *, struct got_object_id *, struct got_object_id *,
+    struct got_pathlist_head *, int, int, int, struct got_repository *, FILE *);
 
 #define GOT_DIFF_MAX_CONTEXT	64
