@@ -50,11 +50,14 @@
 #include "got_lib_object_cache.h"
 #include "got_lib_deflate.h"
 #include "got_lib_pack.h"
+#include "got_lib_pack_create.h"
 #include "got_lib_privsep.h"
 #include "got_lib_repository.h"
 #include "got_lib_ratelimit.h"
 #include "got_lib_inflate.h"
 
+#include "murmurhash2.h"
+
 #ifndef MIN
 #define	MIN(_a,_b) ((_a) < (_b) ? (_a) : (_b))
 #endif
@@ -69,7 +72,7 @@
 
 struct got_pack_meta {
 	struct got_object_id id;
-	char	*path;
+	uint32_t path_hash;
 	int	obj_type;
 	off_t	size;
 	time_t	mtime;
@@ -83,7 +86,6 @@ struct got_pack_meta {
 	off_t	delta_compressed_len; /* encoded+compressed delta length */
 	int	nchain;
 
-	int	have_reused_delta;
 	off_t   reused_delta_offset; /* offset of delta in reused pack file */
 	struct got_object_id *base_obj_id;
 
@@ -102,9 +104,8 @@ struct got_pack_metavec {
 
 static const struct got_error *
 alloc_meta(struct got_pack_meta **new, struct got_object_id *id,
-    const char *path, int obj_type, time_t mtime)
+    const char *path, int obj_type, time_t mtime, uint32_t seed)
 {
-	const struct got_error *err = NULL;
 	struct got_pack_meta *m;
 
 	*new = NULL;
@@ -115,13 +116,7 @@ alloc_meta(struct got_pack_meta **new, struct got_object_id *id,
 
 	memcpy(&m->id, id, sizeof(m->id));
 
-	m->path = strdup(path);
-	if (m->path == NULL) {
-		err = got_error_from_errno("strdup");
-		free(m);
-		return err;
-	}
-
+	m->path_hash = murmurhash2(path, strlen(path), seed);
 	m->obj_type = obj_type;
 	m->mtime = mtime;
 	*new = m;
@@ -133,12 +128,12 @@ clear_meta(struct got_pack_meta *meta)
 {
 	if (meta == NULL)
 		return;
-	free(meta->path);
-	meta->path = NULL;
+	meta->path_hash = 0;
 	free(meta->delta_buf);
 	meta->delta_buf = NULL;
 	free(meta->base_obj_id);
 	meta->base_obj_id = NULL;
+	meta->reused_delta_offset = 0;
 }
 
 static void
@@ -155,16 +150,16 @@ static int
 delta_order_cmp(const void *pa, const void *pb)
 {
 	struct got_pack_meta *a, *b;
-	int cmp;
 
 	a = *(struct got_pack_meta **)pa;
 	b = *(struct got_pack_meta **)pb;
 
 	if (a->obj_type != b->obj_type)
 		return a->obj_type - b->obj_type;
-	cmp = strcmp(a->path, b->path);
-	if (cmp != 0)
-		return cmp;
+	if (a->path_hash < b->path_hash)
+		return -1;
+	if (a->path_hash > b->path_hash)
+		return 1;
 	if (a->mtime < b->mtime)
 		return -1;
 	if (a->mtime > b->mtime)
@@ -590,7 +585,6 @@ recv_reused_delta(struct got_imsg_reused_delta *delta,
 	m->delta_offset = delta->delta_out_offset;
 	m->prev = base;
 	m->size = delta->result_size;
-	m->have_reused_delta = 1;
 	m->reused_delta_offset = delta->delta_offset;
 	m->base_obj_id = got_object_id_dup(&delta->base_id);
 	if (m->base_obj_id == NULL)
@@ -686,8 +680,11 @@ pick_deltas(struct got_pack_meta **meta, int nmeta, int ncolored,
 	off_t size, best_size;
 	const int max_base_candidates = 3;
 	size_t delta_memsize = 0;
-	const size_t max_delta_memsize = 25 * GOT_DELTA_RESULT_SIZE_CACHED_MAX;
+	const size_t max_delta_memsize = 4 * GOT_DELTA_RESULT_SIZE_CACHED_MAX;
 	int outfd = -1;
+	uint32_t delta_seed;
+
+	delta_seed = arc4random();
 
 	qsort(meta, nmeta, sizeof(struct got_pack_meta *), delta_order_cmp);
 	for (i = 0; i < nmeta; i++) {
@@ -714,10 +711,10 @@ pick_deltas(struct got_pack_meta **meta, int nmeta, int ncolored,
 
 		if (raw->f == NULL) {
 			err = got_deltify_init_mem(&m->dtab, raw->data,
-			    raw->hdrlen, raw->size + raw->hdrlen);
+			    raw->hdrlen, raw->size + raw->hdrlen, delta_seed);
 		} else {
 			err = got_deltify_init(&m->dtab, raw->f, raw->hdrlen,
-			    raw->size + raw->hdrlen);
+			    raw->size + raw->hdrlen, delta_seed);
 		}
 		if (err)
 			goto done;
@@ -751,28 +748,28 @@ pick_deltas(struct got_pack_meta **meta, int nmeta, int ncolored,
 			if (raw->f == NULL && base_raw->f == NULL) {
 				err = got_deltify_mem_mem(&deltas, &ndeltas,
 				    raw->data, raw->hdrlen,
-				    raw->size + raw->hdrlen,
+				    raw->size + raw->hdrlen, delta_seed,
 				    base->dtab, base_raw->data,
 				    base_raw->hdrlen,
 				    base_raw->size + base_raw->hdrlen);
 			} else if (raw->f == NULL) {
 				err = got_deltify_mem_file(&deltas, &ndeltas,
 				    raw->data, raw->hdrlen,
-				    raw->size + raw->hdrlen,
+				    raw->size + raw->hdrlen, delta_seed,
 				    base->dtab, base_raw->f,
 				    base_raw->hdrlen,
 				    base_raw->size + base_raw->hdrlen);
 			} else if (base_raw->f == NULL) {
 				err = got_deltify_file_mem(&deltas, &ndeltas,
 				    raw->f, raw->hdrlen,
-				    raw->size + raw->hdrlen,
+				    raw->size + raw->hdrlen, delta_seed,
 				    base->dtab, base_raw->data,
 				    base_raw->hdrlen,
 				    base_raw->size + base_raw->hdrlen);
 			} else {
 				err = got_deltify(&deltas, &ndeltas,
 				    raw->f, raw->hdrlen,
-				    raw->size + raw->hdrlen,
+				    raw->size + raw->hdrlen, delta_seed,
 				    base->dtab, base_raw->f, base_raw->hdrlen,
 				    base_raw->size + base_raw->hdrlen);
 			}
@@ -862,8 +859,8 @@ search_packidx(int *found, struct got_object_id *id,
 static const struct got_error *
 add_object(int want_meta, struct got_object_idset *idset,
     struct got_object_id *id, const char *path, int obj_type,
-    time_t mtime, int loose_obj_only, struct got_repository *repo,
-    int *ncolored, int *nfound, int *ntrees,
+    time_t mtime, uint32_t seed, int loose_obj_only,
+    struct got_repository *repo, int *ncolored, int *nfound, int *ntrees,
     got_pack_progress_cb progress_cb, void *progress_arg,
     struct got_ratelimit *rl)
 {
@@ -880,7 +877,7 @@ add_object(int want_meta, struct got_object_idset *idset,
 	}
 
 	if (want_meta) {
-		err = alloc_meta(&m, id, path, obj_type, mtime);
+		err = alloc_meta(&m, id, path, obj_type, mtime, seed);
 		if (err)
 			return err;
 
@@ -905,21 +902,16 @@ add_object(int want_meta, struct got_object_idset *idset,
 static const struct got_error *
 load_tree_entries(struct got_object_id_queue *ids, int want_meta,
     struct got_object_idset *idset, struct got_object_idset *idset_exclude,
-    struct got_object_id *tree_id,
-    const char *dpath, time_t mtime, struct got_repository *repo,
+    struct got_tree_object *tree,
+    const char *dpath, time_t mtime, uint32_t seed, struct got_repository *repo,
     int loose_obj_only, int *ncolored, int *nfound, int *ntrees,
     got_pack_progress_cb progress_cb, void *progress_arg,
     struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err;
-	struct got_tree_object *tree;
 	char *p = NULL;
 	int i;
 
-	err = got_object_open_as_tree(&tree, repo, tree_id);
-	if (err)
-		return err;
-
 	(*ntrees)++;
 	err = report_progress(progress_cb, progress_arg, rl,
 	    *ncolored, *nfound, *ntrees, 0L, 0, 0, 0, 0);
@@ -941,8 +933,16 @@ load_tree_entries(struct got_object_id_queue *ids, int want_meta,
 		    got_object_idset_contains(idset, id) ||
 		    got_object_idset_contains(idset_exclude, id))
 			continue;
-		
-		if (asprintf(&p, "%s%s%s", dpath, dpath[0] != '\0' ? "/" : "",
+
+		/*
+		 * If got-read-pack is crawling trees for us then
+		 * we are only here to collect blob IDs.
+		 */
+		if (ids == NULL && S_ISDIR(mode))
+			continue;
+
+		if (asprintf(&p, "%s%s%s", dpath,
+		    got_path_is_root_dir(dpath) ? "" : "/",
 		    got_tree_entry_get_name(e)) == -1) {
 			err = got_error_from_errno("asprintf");
 			break;
@@ -953,21 +953,25 @@ load_tree_entries(struct got_object_id_queue *ids, int want_meta,
 			err = got_object_qid_alloc(&qid, id);
 			if (err)
 				break;
+			qid->data = p;
+			p = NULL;
 			STAILQ_INSERT_TAIL(ids, qid, entry);
 		} else if (S_ISREG(mode) || S_ISLNK(mode)) {
 			err = add_object(want_meta,
 			    want_meta ? idset : idset_exclude, id, p,
-			    GOT_OBJ_TYPE_BLOB, mtime, loose_obj_only, repo,
-			    ncolored, nfound, ntrees,
+			    GOT_OBJ_TYPE_BLOB, mtime, seed, loose_obj_only,
+			    repo, ncolored, nfound, ntrees,
 			    progress_cb, progress_arg, rl);
 			if (err)
 				break;
+			free(p);
+			p = NULL;
+		} else {
+			free(p);
+			p = NULL;
 		}
-		free(p);
-		p = NULL;
 	}
 
-	got_object_tree_close(tree);
 	free(p);
 	return err;
 }
@@ -976,7 +980,7 @@ static const struct got_error *
 load_tree(int want_meta, struct got_object_idset *idset,
     struct got_object_idset *idset_exclude,
     struct got_object_id *tree_id, const char *dpath, time_t mtime,
-    struct got_repository *repo, int loose_obj_only,
+    uint32_t seed, struct got_repository *repo, int loose_obj_only,
     int *ncolored, int *nfound, int *ntrees,
     got_pack_progress_cb progress_cb, void *progress_arg,
     struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
@@ -984,6 +988,7 @@ load_tree(int want_meta, struct got_object_idset *idset,
 	const struct got_error *err = NULL;
 	struct got_object_id_queue tree_ids;
 	struct got_object_qid *qid;
+	struct got_tree_object *tree = NULL;
 
 	if (got_object_idset_contains(idset, tree_id) ||
 	    got_object_idset_contains(idset_exclude, tree_id))
@@ -992,11 +997,18 @@ load_tree(int want_meta, struct got_object_idset *idset,
 	err = got_object_qid_alloc(&qid, tree_id);
 	if (err)
 		return err;
+	qid->data = strdup(dpath);
+	if (qid->data == NULL) {
+		err = got_error_from_errno("strdup");
+		got_object_qid_free(qid);
+		return err;
+	}
 
 	STAILQ_INIT(&tree_ids);
 	STAILQ_INSERT_TAIL(&tree_ids, qid, entry);
 
 	while (!STAILQ_EMPTY(&tree_ids)) {
+		const char *path;
 		if (cancel_cb) {
 			err = (*cancel_cb)(cancel_arg);
 			if (err)
@@ -1005,41 +1017,59 @@ load_tree(int want_meta, struct got_object_idset *idset,
 
 		qid = STAILQ_FIRST(&tree_ids);
 		STAILQ_REMOVE_HEAD(&tree_ids, entry);
+		path = qid->data;
 
 		if (got_object_idset_contains(idset, &qid->id) ||
 		    got_object_idset_contains(idset_exclude, &qid->id)) {
+			free(qid->data);
 			got_object_qid_free(qid);
 			continue;
 		}
 
 		err = add_object(want_meta, want_meta ? idset : idset_exclude,
-		    &qid->id, dpath, GOT_OBJ_TYPE_TREE,
-		    mtime, loose_obj_only, repo,
+		    &qid->id, path, GOT_OBJ_TYPE_TREE,
+		    mtime, seed, loose_obj_only, repo,
 		    ncolored, nfound, ntrees, progress_cb, progress_arg, rl);
 		if (err) {
+			free(qid->data);
+			got_object_qid_free(qid);
+			break;
+		}
+
+		err = got_object_open_as_tree(&tree, repo, &qid->id);
+		if (err) {
+			free(qid->data);
 			got_object_qid_free(qid);
 			break;
 		}
 
 		err = load_tree_entries(&tree_ids, want_meta, idset,
-		    idset_exclude, &qid->id,
-		    dpath, mtime, repo, loose_obj_only, ncolored, nfound,
-		    ntrees, progress_cb, progress_arg, rl,
+		    idset_exclude, tree, path, mtime, seed, repo,
+		    loose_obj_only, ncolored, nfound, ntrees,
+		    progress_cb, progress_arg, rl,
 		    cancel_cb, cancel_arg);
+		free(qid->data);
 		got_object_qid_free(qid);
 		if (err)
 			break;
+
+		got_object_tree_close(tree);
+		tree = NULL;
 	}
 
+	STAILQ_FOREACH(qid, &tree_ids, entry)
+		free(qid->data);
 	got_object_id_queue_free(&tree_ids);
+	if (tree)
+		got_object_tree_close(tree);
 	return err;
 }
 
 static const struct got_error *
 load_commit(int want_meta, struct got_object_idset *idset,
     struct got_object_idset *idset_exclude,
-    struct got_object_id *id, struct got_repository *repo, int loose_obj_only,
-    int *ncolored, int *nfound, int *ntrees,
+    struct got_object_id *id, struct got_repository *repo, uint32_t seed,
+    int loose_obj_only, int *ncolored, int *nfound, int *ntrees,
     got_pack_progress_cb progress_cb, void *progress_arg,
     struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
 {
@@ -1065,7 +1095,7 @@ load_commit(int want_meta, struct got_object_idset *idset,
 
 	err = add_object(want_meta, want_meta ? idset : idset_exclude,
 	    id, "", GOT_OBJ_TYPE_COMMIT,
-	    got_object_commit_get_committer_time(commit),
+	    got_object_commit_get_committer_time(commit), seed,
 	    loose_obj_only, repo,
 	    ncolored, nfound, ntrees, progress_cb, progress_arg, rl);
 	if (err)
@@ -1073,7 +1103,7 @@ load_commit(int want_meta, struct got_object_idset *idset,
 
 	err = load_tree(want_meta, idset, idset_exclude,
 	    got_object_commit_get_tree_id(commit),
-	    "", got_object_commit_get_committer_time(commit),
+	    "", got_object_commit_get_committer_time(commit), seed,
 	    repo, loose_obj_only, ncolored, nfound, ntrees,
 	    progress_cb, progress_arg, rl, cancel_cb, cancel_arg);
 done:
@@ -1084,8 +1114,8 @@ done:
 static const struct got_error *
 load_tag(int want_meta, struct got_object_idset *idset,
     struct got_object_idset *idset_exclude,
-    struct got_object_id *id, struct got_repository *repo, int loose_obj_only,
-    int *ncolored, int *nfound, int *ntrees,
+    struct got_object_id *id, struct got_repository *repo, uint32_t seed,
+    int loose_obj_only, int *ncolored, int *nfound, int *ntrees,
     got_pack_progress_cb progress_cb, void *progress_arg,
     struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
 {
@@ -1111,7 +1141,7 @@ load_tag(int want_meta, struct got_object_idset *idset,
 
 	err = add_object(want_meta, want_meta ? idset : idset_exclude,
 	    id, "", GOT_OBJ_TYPE_TAG,
-	    got_object_tag_get_tagger_time(tag), loose_obj_only, repo,
+	    got_object_tag_get_tagger_time(tag), seed, loose_obj_only, repo,
 	    ncolored, nfound, ntrees, progress_cb, progress_arg, rl);
 	if (err)
 		goto done;
@@ -1119,16 +1149,16 @@ load_tag(int want_meta, struct got_object_idset *idset,
 	switch (got_object_tag_get_object_type(tag)) {
 	case GOT_OBJ_TYPE_COMMIT:
 		err = load_commit(want_meta, idset, idset_exclude,
-		    got_object_tag_get_object_id(tag), repo, loose_obj_only,
-		    ncolored, nfound, ntrees, progress_cb, progress_arg, rl,
-		    cancel_cb, cancel_arg);
+		    got_object_tag_get_object_id(tag), repo, seed,
+		    loose_obj_only, ncolored, nfound, ntrees,
+		    progress_cb, progress_arg, rl, cancel_cb, cancel_arg);
 		break;
 	case GOT_OBJ_TYPE_TREE:
 		err = load_tree(want_meta, idset, idset_exclude,
 		    got_object_tag_get_object_id(tag), "",
-		    got_object_tag_get_tagger_time(tag), repo, loose_obj_only,
-		    ncolored, nfound, ntrees, progress_cb, progress_arg, rl,
-		    cancel_cb, cancel_arg);
+		    got_object_tag_get_tagger_time(tag), seed, repo,
+		    loose_obj_only, ncolored, nfound, ntrees,
+		    progress_cb, progress_arg, rl, cancel_cb, cancel_arg);
 		break;
 	default:
 		break;
@@ -1276,7 +1306,6 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				err = paint_commit(qid, COLOR_SKIP);
 				if (err)
 					goto done;
-				nskip++;
 			} else
 				(*ncolored)++;
 			err = got_object_idset_add(keep, &qid->id, NULL);
@@ -1292,7 +1321,6 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				err = paint_commit(qid, COLOR_SKIP);
 				if (err)
 					goto done;
-				nskip++;
 			} else
 				(*ncolored)++;
 			err = got_object_idset_add(drop, &qid->id, NULL);
@@ -1435,16 +1463,219 @@ done:
 	return err;
 }
 
+struct load_packed_obj_arg {
+	/* output parameters: */
+	struct got_object_id *id;
+	char *dpath;
+	time_t mtime;
+
+	/* input parameters: */
+	uint32_t seed;
+	int want_meta;
+	struct got_object_idset *idset;
+	struct got_object_idset *idset_exclude;
+	int loose_obj_only;
+	int *ncolored;
+	int *nfound;
+	int *ntrees;
+	got_pack_progress_cb progress_cb;
+	void *progress_arg;
+	struct got_ratelimit *rl;
+	got_cancel_cb cancel_cb;
+	void *cancel_arg;
+};
+
+static const struct got_error *
+load_packed_commit_id(void *arg, time_t mtime, struct got_object_id *id,
+    struct got_repository *repo)
+{
+	struct load_packed_obj_arg *a = arg;
+
+	if (got_object_idset_contains(a->idset, id) ||
+	    got_object_idset_contains(a->idset_exclude, id))
+		return NULL;
+
+	return add_object(a->want_meta,
+	    a->want_meta ? a->idset : a->idset_exclude,
+	    id, "", GOT_OBJ_TYPE_COMMIT, mtime, a->seed, a->loose_obj_only,
+	    repo, a->ncolored, a->nfound, a->ntrees,
+	    a->progress_cb, a->progress_arg, a->rl);
+}
+
+static const struct got_error *
+load_packed_tree_ids(void *arg, struct got_tree_object *tree, time_t mtime,
+    struct got_object_id *id, const char *dpath, struct got_repository *repo)
+{
+	const struct got_error *err;
+	struct load_packed_obj_arg *a = arg;
+	const char *relpath;
+
+	/*
+	 * When we receive a tree's ID and path but not the tree itself,
+	 * this tree object was not found in the pack file. This is the
+	 * last time we are being called for this optimized traversal.
+	 * Return from here and switch to loading objects the slow way.
+	 */
+	if (tree == NULL) {
+		free(a->id);
+		a->id = got_object_id_dup(id);
+		if (a->id == NULL) {
+			err = got_error_from_errno("got_object_id_dup");
+			free(a->dpath);
+			a->dpath = NULL;
+			return err;
+		}
+
+		free(a->dpath);
+		a->dpath = strdup(dpath);
+		if (a->dpath == NULL) {
+			err = got_error_from_errno("strdup");
+			free(a->id);
+			a->id = NULL;
+			return err;
+		}
+
+		a->mtime = mtime;
+		return NULL;
+	}
+
+	if (got_object_idset_contains(a->idset, id) ||
+	    got_object_idset_contains(a->idset_exclude, id))
+		return NULL;
+
+	relpath = dpath;
+	while (relpath[0] == '/')
+		relpath++;
+
+	err = add_object(a->want_meta,
+	    a->want_meta ? a->idset : a->idset_exclude,
+	    id, relpath, GOT_OBJ_TYPE_TREE, mtime, a->seed,
+	    a->loose_obj_only, repo, a->ncolored, a->nfound, a->ntrees,
+	    a->progress_cb, a->progress_arg, a->rl);
+	if (err)
+		return err;
+
+	return load_tree_entries(NULL, a->want_meta, a->idset,
+	    a->idset_exclude, tree, dpath, mtime, a->seed, repo,
+	    a->loose_obj_only, a->ncolored, a->nfound, a->ntrees,
+	    a->progress_cb, a->progress_arg, a->rl,
+	    a->cancel_cb, a->cancel_arg);
+}
+
+static const struct got_error *
+load_packed_object_ids(int *found_all_objects,
+    struct got_object_id **ours, int nours,
+    struct got_object_id **theirs, int ntheirs,
+    int want_meta, uint32_t seed, struct got_object_idset *idset,
+    struct got_object_idset *idset_exclude, int loose_obj_only,
+    struct got_repository *repo, struct got_packidx *packidx,
+    int *ncolored, int *nfound, int *ntrees,
+    got_pack_progress_cb progress_cb, void *progress_arg,
+    struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
+{
+	const struct got_error *err = NULL;
+	struct load_packed_obj_arg lpa;
+
+	memset(&lpa, 0, sizeof(lpa));
+	lpa.seed = seed;
+	lpa.want_meta = want_meta;
+	lpa.idset = idset;
+	lpa.idset_exclude = idset_exclude;
+	lpa.loose_obj_only = loose_obj_only;
+	lpa.ncolored = ncolored;
+	lpa.nfound = nfound;
+	lpa.ntrees = ntrees;
+	lpa.progress_cb = progress_cb;
+	lpa.progress_arg = progress_arg;
+	lpa.rl = rl;
+	lpa.cancel_cb = cancel_cb;
+	lpa.cancel_arg = cancel_arg;
+
+	/* Attempt to load objects via got-read-pack, as far as possible. */
+	err = got_object_enumerate(found_all_objects, load_packed_commit_id,
+	   load_packed_tree_ids, &lpa, ours, nours, theirs, ntheirs,
+	   packidx, repo);
+	if (err)
+		return err;
+
+	if (lpa.id == NULL)
+		return NULL;
+
+	/*
+	 * An incomplete tree hierarchy was present in the pack file
+	 * and caused loading to be aborted.
+	 * Continue loading trees the slow way.
+	 */
+	err = load_tree(want_meta, idset, idset_exclude,
+	    lpa.id, lpa.dpath, lpa.mtime, seed, repo, loose_obj_only,
+	    ncolored, nfound, ntrees, progress_cb, progress_arg, rl,
+	    cancel_cb, cancel_arg);
+	free(lpa.id);
+	free(lpa.dpath);
+	return err;
+}
+
+static const struct got_error *
+find_pack_for_enumeration(struct got_packidx **best_packidx,
+    struct got_object_id **ids, int nids, struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	struct got_pathlist_entry *pe;
+	const char *best_packidx_path = NULL;
+	int nobj_max = 0;
+	int ncommits_max = 0;
+
+	*best_packidx = NULL;
+
+	/*
+	 * Find the largest pack which contains at least some of the
+	 * commits and tags we are interested in.
+	 */
+	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+		const char *path_packidx = pe->path;
+		struct got_packidx *packidx;
+		int nobj, i, idx, ncommits = 0;
+
+		err = got_repo_get_packidx(&packidx, path_packidx, repo);
+		if (err)
+			break;
+
+		nobj = be32toh(packidx->hdr.fanout_table[0xff]);
+		if (nobj <= nobj_max)
+			continue;
+
+		for (i = 0; i < nids; i++) {
+			idx = got_packidx_get_object_idx(packidx, ids[i]);
+			if (idx != -1)
+				ncommits++;
+		}
+		if (ncommits > ncommits_max) {
+			best_packidx_path = path_packidx;
+			nobj_max = nobj;
+			ncommits_max = ncommits;
+		}
+	}
+
+	if (best_packidx_path && err == NULL) {
+		err = got_repo_get_packidx(best_packidx, best_packidx_path,
+		    repo);
+	}
+
+	return err;
+}
+
 static const struct got_error *
 load_object_ids(int *ncolored, int *nfound, int *ntrees,
     struct got_object_idset *idset, struct got_object_id **theirs, int ntheirs,
     struct got_object_id **ours, int nours, struct got_repository *repo,
-    int loose_obj_only, got_pack_progress_cb progress_cb, void *progress_arg,
-    struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
+    uint32_t seed, int loose_obj_only, got_pack_progress_cb progress_cb,
+    void *progress_arg, struct got_ratelimit *rl, got_cancel_cb cancel_cb,
+    void *cancel_arg)
 {
 	const struct got_error *err = NULL;
 	struct got_object_id **ids = NULL;
-	int i, nobj = 0, obj_type;
+	struct got_packidx *packidx = NULL;
+	int i, nobj = 0, obj_type, found_all_objects = 0;
 	struct got_object_idset *idset_exclude;
 
 	idset_exclude = got_object_idset_alloc();
@@ -1460,6 +1691,18 @@ load_object_ids(int *ncolored, int *nfound, int *ntrees,
 	if (err)
 		goto done;
 
+	err = find_pack_for_enumeration(&packidx, theirs, ntheirs, repo);
+	if (err)
+		goto done;
+	if (packidx) {
+		err = load_packed_object_ids(&found_all_objects,
+		    theirs, ntheirs, NULL, 0, 0, seed, idset, idset_exclude,
+		    loose_obj_only, repo, packidx, ncolored, nfound, ntrees,
+		    progress_cb, progress_arg, rl, cancel_cb, cancel_arg);
+		if (err)
+			goto done;
+	}
+
 	for (i = 0; i < ntheirs; i++) {
 		struct got_object_id *id = theirs[i];
 		if (id == NULL)
@@ -1468,15 +1711,18 @@ load_object_ids(int *ncolored, int *nfound, int *ntrees,
 		if (err)
 			return err;
 		if (obj_type == GOT_OBJ_TYPE_COMMIT) {
-			err = load_commit(0, idset, idset_exclude, id, repo,
-			    loose_obj_only, ncolored, nfound, ntrees,
-			    progress_cb, progress_arg, rl,
-			    cancel_cb, cancel_arg);
-			if (err)
-				goto done;
+			if (!found_all_objects) {
+				err = load_commit(0, idset, idset_exclude,
+				    id, repo, seed, loose_obj_only,
+				    ncolored, nfound, ntrees,
+				    progress_cb, progress_arg, rl,
+				    cancel_cb, cancel_arg);
+				if (err)
+					goto done;
+			}
 		} else if (obj_type == GOT_OBJ_TYPE_TAG) {
 			err = load_tag(0, idset, idset_exclude, id, repo,
-			    loose_obj_only, ncolored, nfound, ntrees,
+			    seed, loose_obj_only, ncolored, nfound, ntrees,
 			    progress_cb, progress_arg, rl,
 			    cancel_cb, cancel_arg);
 			if (err)
@@ -1484,14 +1730,30 @@ load_object_ids(int *ncolored, int *nfound, int *ntrees,
 		}
 	}
 
-	for (i = 0; i < nobj; i++) {
-		err = load_commit(1, idset, idset_exclude,
-		    ids[i], repo, loose_obj_only, ncolored, nfound, ntrees,
+	found_all_objects = 0;
+	err = find_pack_for_enumeration(&packidx, ids, nobj, repo);
+	if (err)
+		goto done;
+	if (packidx) {
+		err = load_packed_object_ids(&found_all_objects, ids,
+		    nobj, theirs, ntheirs, 1, seed, idset, idset_exclude,
+		    loose_obj_only, repo, packidx, ncolored, nfound, ntrees,
 		    progress_cb, progress_arg, rl, cancel_cb, cancel_arg);
 		if (err)
 			goto done;
 	}
 
+	if (!found_all_objects) {
+		for (i = 0; i < nobj; i++) {
+			err = load_commit(1, idset, idset_exclude, ids[i],
+			    repo, seed, loose_obj_only, ncolored, nfound,
+			    ntrees, progress_cb, progress_arg, rl,
+			    cancel_cb, cancel_arg);
+			if (err)
+				goto done;
+		}
+	}
+
 	for (i = 0; i < nours; i++) {
 		struct got_object_id *id = ours[i];
 		struct got_pack_meta *m;
@@ -1507,7 +1769,7 @@ load_object_ids(int *ncolored, int *nfound, int *ntrees,
 		if (obj_type != GOT_OBJ_TYPE_TAG)
 			continue;
 		err = load_tag(1, idset, idset_exclude, id, repo,
-		    loose_obj_only, ncolored, nfound, ntrees,
+		    seed, loose_obj_only, ncolored, nfound, ntrees,
 		    progress_cb, progress_arg, rl, cancel_cb, cancel_arg);
 		if (err)
 			goto done;
@@ -1909,7 +2171,7 @@ add_meta_idset_cb(struct got_object_id *id, void *data, void *arg)
 	struct got_pack_meta *m = data;
 	struct got_pack_metavec *v = arg;
 
-	if (m->have_reused_delta)
+	if (m->reused_delta_offset != 0)
 		return NULL;
 
 	return add_meta(m, v);
@@ -1931,6 +2193,9 @@ got_pack_create(uint8_t *packsha1, FILE *packfile,
 	struct got_pack_metavec deltify, reuse;
 	int ncolored = 0, nfound = 0, ntrees = 0;
 	size_t ndeltify;
+	uint32_t seed;
+
+	seed = arc4random();
 
 	memset(&deltify, 0, sizeof(deltify));
 	memset(&reuse, 0, sizeof(reuse));
@@ -1942,10 +2207,10 @@ got_pack_create(uint8_t *packsha1, FILE *packfile,
 		return got_error_from_errno("got_object_idset_alloc");
 
 	err = load_object_ids(&ncolored, &nfound, &ntrees, idset, theirs,
-	    ntheirs, ours, nours, repo, loose_obj_only,
+	    ntheirs, ours, nours, repo, seed, loose_obj_only,
 	    progress_cb, progress_arg, &rl, cancel_cb, cancel_arg);
 	if (err)
-		return err;
+		goto done;
 
 	if (progress_cb) {
 		err = progress_cb(progress_arg, ncolored, nfound, ntrees,
