@@ -296,6 +296,7 @@ got_privsep_send_raw_obj(struct imsgbuf *ibuf, off_t size, size_t hdrlen,
 	size_t len = sizeof(iobj);
 	struct ibuf *wbuf;
 
+	memset(&iobj, 0, sizeof(iobj));
 	iobj.hdrlen = hdrlen;
 	iobj.size = size;
 
@@ -308,18 +309,12 @@ got_privsep_send_raw_obj(struct imsgbuf *ibuf, off_t size, size_t hdrlen,
 		return err;
 	}
 
-	if (imsg_add(wbuf, &iobj, sizeof(iobj)) == -1) {
-		err = got_error_from_errno("imsg_add RAW_OBJECT");
-		ibuf_free(wbuf);
-		return err;
-	}
+	if (imsg_add(wbuf, &iobj, sizeof(iobj)) == -1)
+		return got_error_from_errno("imsg_add RAW_OBJECT");
 
 	if (data && size + hdrlen <= GOT_PRIVSEP_INLINE_OBJECT_DATA_MAX) {
-		if (imsg_add(wbuf, data, size + hdrlen) == -1) {
-			err = got_error_from_errno("imsg_add RAW_OBJECT");
-			ibuf_free(wbuf);
-			return err;
-		}
+		if (imsg_add(wbuf, data, size + hdrlen) == -1)
+			return got_error_from_errno("imsg_add RAW_OBJECT");
 	}
 
 	wbuf->fd = -1;
@@ -360,7 +355,8 @@ got_privsep_recv_raw_obj(uint8_t **outbuf, off_t *size, size_t *hdrlen,
 			break;
 		}
 
-		if (*size + *hdrlen > GOT_PRIVSEP_INLINE_OBJECT_DATA_MAX) {
+		if (*size < 0 ||
+		    *size + *hdrlen > GOT_PRIVSEP_INLINE_OBJECT_DATA_MAX) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
 			break;
 		}
@@ -391,6 +387,7 @@ got_privsep_send_commit_req(struct imsgbuf *ibuf, int fd,
 	void *data;
 	size_t len;
 
+	memset(&iobj, 0, sizeof(iobj));
 	if (pack_idx != -1) { /* commit is packed */
 		iobj.idx = pack_idx;
 		memcpy(iobj.id, id->sha1, sizeof(iobj.id));
@@ -415,7 +412,6 @@ const struct got_error *
 got_privsep_send_tree_req(struct imsgbuf *ibuf, int fd,
     struct got_object_id *id, int pack_idx)
 {
-	const struct got_error *err = NULL;
 	struct ibuf *wbuf;
 	size_t len;
 
@@ -428,18 +424,12 @@ got_privsep_send_tree_req(struct imsgbuf *ibuf, int fd,
 	if (wbuf == NULL)
 		return got_error_from_errno("imsg_create TREE_REQUEST");
 
-	if (imsg_add(wbuf, id->sha1, SHA1_DIGEST_LENGTH) == -1) {
-		err = got_error_from_errno("imsg_add TREE_REQUEST");
-		ibuf_free(wbuf);
-		return err;
-	}
+	if (imsg_add(wbuf, id->sha1, SHA1_DIGEST_LENGTH) == -1)
+		return got_error_from_errno("imsg_add TREE_REQUEST");
 
 	if (pack_idx != -1) { /* tree is packed */
-		if (imsg_add(wbuf, &pack_idx, sizeof(pack_idx)) == -1) {
-			err = got_error_from_errno("imsg_add TREE_REQUEST");
-			ibuf_free(wbuf);
-			return err;
-		}
+		if (imsg_add(wbuf, &pack_idx, sizeof(pack_idx)) == -1)
+			return got_error_from_errno("imsg_add TREE_REQUEST");
 	}
 
 	wbuf->fd = fd;
@@ -456,6 +446,7 @@ got_privsep_send_tag_req(struct imsgbuf *ibuf, int fd,
 	void *data;
 	size_t len;
 
+	memset(&iobj, 0, sizeof(iobj));
 	if (pack_idx != -1) { /* tag is packed */
 		iobj.idx = pack_idx;
 		memcpy(iobj.id, id->sha1, sizeof(iobj.id));
@@ -482,6 +473,7 @@ got_privsep_send_blob_req(struct imsgbuf *ibuf, int infd,
 	void *data;
 	size_t len;
 
+	memset(&iobj, 0, sizeof(iobj));
 	if (pack_idx != -1) { /* blob is packed */
 		iobj.idx = pack_idx;
 		memcpy(iobj.id, id->sha1, sizeof(iobj.id));
@@ -542,6 +534,8 @@ got_privsep_send_obj(struct imsgbuf *ibuf, struct got_object *obj)
 {
 	struct got_imsg_object iobj;
 
+	memset(&iobj, 0, sizeof(iobj));
+
 	memcpy(iobj.id, obj->id.sha1, sizeof(iobj.id));
 	iobj.type = obj->type;
 	iobj.flags = obj->flags;
@@ -610,21 +604,12 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 			return got_error_from_errno("imsg_create FETCH_HAVE_REF");
 
 		/* Keep in sync with struct got_imsg_fetch_have_ref! */
-		if (imsg_add(wbuf, id->sha1, sizeof(id->sha1)) == -1) {
-			err = got_error_from_errno("imsg_add FETCH_HAVE_REF");
-			ibuf_free(wbuf);
-			return err;
-		}
-		if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1) {
-			err = got_error_from_errno("imsg_add FETCH_HAVE_REF");
-			ibuf_free(wbuf);
-			return err;
-		}
-		if (imsg_add(wbuf, name, name_len) == -1) {
-			err = got_error_from_errno("imsg_add FETCH_HAVE_REF");
-			ibuf_free(wbuf);
-			return err;
-		}
+		if (imsg_add(wbuf, id->sha1, sizeof(id->sha1)) == -1)
+			return got_error_from_errno("imsg_add FETCH_HAVE_REF");
+		if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1)
+			return got_error_from_errno("imsg_add FETCH_HAVE_REF");
+		if (imsg_add(wbuf, name, name_len) == -1)
+			return got_error_from_errno("imsg_add FETCH_HAVE_REF");
 
 		wbuf->fd = -1;
 		imsg_close(ibuf, wbuf);
@@ -645,18 +630,12 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 			    "imsg_create FETCH_WANTED_BRANCH");
 
 		/* Keep in sync with struct got_imsg_fetch_wanted_branch! */
-		if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1) {
-			err = got_error_from_errno(
+		if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1)
+			return got_error_from_errno(
 			    "imsg_add FETCH_WANTED_BRANCH");
-			ibuf_free(wbuf);
-			return err;
-		}
-		if (imsg_add(wbuf, name, name_len) == -1) {
-			err = got_error_from_errno(
+		if (imsg_add(wbuf, name, name_len) == -1)
+			return got_error_from_errno(
 			    "imsg_add FETCH_WANTED_BRANCH");
-			ibuf_free(wbuf);
-			return err;
-		}
 
 		wbuf->fd = -1;
 		imsg_close(ibuf, wbuf);
@@ -677,18 +656,12 @@ got_privsep_send_fetch_req(struct imsgbuf *ibuf, int fd,
 			    "imsg_create FETCH_WANTED_REF");
 
 		/* Keep in sync with struct got_imsg_fetch_wanted_ref! */
-		if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1) {
-			err = got_error_from_errno(
+		if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1)
+			return got_error_from_errno(
 			    "imsg_add FETCH_WANTED_REF");
-			ibuf_free(wbuf);
-			return err;
-		}
-		if (imsg_add(wbuf, name, name_len) == -1) {
-			err = got_error_from_errno(
+		if (imsg_add(wbuf, name, name_len) == -1)
+			return got_error_from_errno(
 			    "imsg_add FETCH_WANTED_REF");
-			ibuf_free(wbuf);
-			return err;
-		}
 
 		wbuf->fd = -1;
 		imsg_close(ibuf, wbuf);
@@ -735,10 +708,6 @@ got_privsep_recv_fetch_progress(int *done, struct got_object_id **id,
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 	switch (imsg.hdr.type) {
 	case GOT_IMSG_ERROR:
-		if (datalen < sizeof(struct got_imsg_error)) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
-			break;
-		}
 		err = recv_imsg_error(&imsg, datalen);
 		break;
 	case GOT_IMSG_FETCH_SYMREFS:
@@ -867,7 +836,6 @@ static const struct got_error *
 send_send_ref(const char *name, size_t name_len, struct got_object_id *id,
     int delete, struct imsgbuf *ibuf)
 {
-	const struct got_error *err = NULL;
 	size_t len;
 	struct ibuf *wbuf;
 
@@ -877,26 +845,14 @@ send_send_ref(const char *name, size_t name_len, struct got_object_id *id,
 		return got_error_from_errno("imsg_create SEND_REF");
 
 	/* Keep in sync with struct got_imsg_send_ref! */
-	if (imsg_add(wbuf, id->sha1, sizeof(id->sha1)) == -1) {
-		err = got_error_from_errno("imsg_add SEND_REF");
-		ibuf_free(wbuf);
-		return err;
-	}
-	if (imsg_add(wbuf, &delete, sizeof(delete)) == -1) {
-		err = got_error_from_errno("imsg_add SEND_REF");
-		ibuf_free(wbuf);
-		return err;
-	}
-	if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1) {
-		err = got_error_from_errno("imsg_add SEND_REF");
-		ibuf_free(wbuf);
-		return err;
-	}
-	if (imsg_add(wbuf, name, name_len) == -1) {
-		err = got_error_from_errno("imsg_add SEND_REF");
-		ibuf_free(wbuf);
-		return err;
-	}
+	if (imsg_add(wbuf, id->sha1, sizeof(id->sha1)) == -1)
+		return got_error_from_errno("imsg_add SEND_REF");
+	if (imsg_add(wbuf, &delete, sizeof(delete)) == -1)
+		return got_error_from_errno("imsg_add SEND_REF");
+	if (imsg_add(wbuf, &name_len, sizeof(name_len)) == -1)
+		return got_error_from_errno("imsg_add SEND_REF");
+	if (imsg_add(wbuf, name, name_len) == -1)
+		return got_error_from_errno("imsg_add SEND_REF");
 
 	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
@@ -976,10 +932,6 @@ got_privsep_recv_send_remote_refs(struct got_pathlist_head *remote_refs,
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 		switch (imsg.hdr.type) {
 		case GOT_IMSG_ERROR:
-			if (datalen < sizeof(struct got_imsg_error)) {
-				err = got_error(GOT_ERR_PRIVSEP_LEN);
-				goto done;
-			}
 			err = recv_imsg_error(&imsg, datalen);
 			goto done;
 		case GOT_IMSG_SEND_REMOTE_REF:
@@ -1063,10 +1015,6 @@ got_privsep_recv_send_progress(int *done, off_t *bytes_sent,
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 	switch (imsg.hdr.type) {
 	case GOT_IMSG_ERROR:
-		if (datalen < sizeof(struct got_imsg_error)) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
-			break;
-		}
 		err = recv_imsg_error(&imsg, datalen);
 		break;
 	case GOT_IMSG_SEND_UPLOAD_PROGRESS:
@@ -1150,10 +1098,6 @@ got_privsep_recv_index_progress(int *done, int *nobj_total,
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 	switch (imsg.hdr.type) {
 	case GOT_IMSG_ERROR:
-		if (datalen < sizeof(struct got_imsg_error)) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
-			break;
-		}
 		err = recv_imsg_error(&imsg, datalen);
 		break;
 	case GOT_IMSG_IDXPACK_PROGRESS:
@@ -1162,6 +1106,11 @@ got_privsep_recv_index_progress(int *done, int *nobj_total,
 			break;
 		}
 		iprogress = (struct got_imsg_index_pack_progress *)imsg.data;
+		if (iprogress->nobj_total < 0 || iprogress->nobj_indexed < 0 ||
+		    iprogress->nobj_loose < 0 || iprogress->nobj_resolved < 0) {
+			err = got_error(GOT_ERR_RANGE);
+			break;
+		}
 		*nobj_total = iprogress->nobj_total;
 		*nobj_indexed = iprogress->nobj_indexed;
 		*nobj_loose = iprogress->nobj_loose;
@@ -1194,6 +1143,9 @@ got_privsep_get_imsg_obj(struct got_object **obj, struct imsg *imsg,
 		return got_error(GOT_ERR_PRIVSEP_LEN);
 	iobj = imsg->data;
 
+	if (iobj->pack_offset < 0)
+		return got_error(GOT_ERR_PACK_OFFSET);
+
 	*obj = calloc(1, sizeof(**obj));
 	if (*obj == NULL)
 		return got_error_from_errno("calloc");
@@ -1363,40 +1315,18 @@ get_commit_from_imsg(struct got_commit_object **commit,
 	(*commit)->committer_time = icommit->committer_time;
 	(*commit)->committer_gmtoff = icommit->committer_gmtoff;
 
-	if (icommit->author_len == 0) {
-		(*commit)->author = strdup("");
-		if ((*commit)->author == NULL) {
-			err = got_error_from_errno("strdup");
-			goto done;
-		}
-	} else {
-		(*commit)->author = malloc(icommit->author_len + 1);
-		if ((*commit)->author == NULL) {
-			err = got_error_from_errno("malloc");
-			goto done;
-		}
-		memcpy((*commit)->author, imsg->data + len,
-		    icommit->author_len);
-		(*commit)->author[icommit->author_len] = '\0';
+	(*commit)->author = strndup(imsg->data + len, icommit->author_len);
+	if ((*commit)->author == NULL) {
+		err = got_error_from_errno("strndup");
+		goto done;
 	}
 	len += icommit->author_len;
 
-	if (icommit->committer_len == 0) {
-		(*commit)->committer = strdup("");
-		if ((*commit)->committer == NULL) {
-			err = got_error_from_errno("strdup");
-			goto done;
-		}
-	} else {
-		(*commit)->committer =
-		    malloc(icommit->committer_len + 1);
-		if ((*commit)->committer == NULL) {
-			err = got_error_from_errno("malloc");
-			goto done;
-		}
-		memcpy((*commit)->committer, imsg->data + len,
-		    icommit->committer_len);
-		(*commit)->committer[icommit->committer_len] = '\0';
+	(*commit)->committer = strndup(imsg->data + len,
+	    icommit->committer_len);
+	if ((*commit)->committer == NULL) {
+		err = got_error_from_errno("strndup");
+		goto done;
 	}
 	len += icommit->committer_len;
 
@@ -1488,72 +1418,164 @@ got_privsep_recv_commit(struct got_commit_object **commit, struct imsgbuf *ibuf)
 	return err;
 }
 
-const struct got_error *
-got_privsep_send_tree(struct imsgbuf *ibuf, struct got_pathlist_head *entries,
+static const struct got_error *
+send_tree_entries_batch(struct imsgbuf *ibuf,
+    struct got_parsed_tree_entry *entries, int idx0, int idxN, size_t len)
+{
+	struct ibuf *wbuf;
+	struct got_imsg_tree_entries ientries;
+	int i;
+
+	memset(&ientries, 0, sizeof(ientries));
+
+	wbuf = imsg_create(ibuf, GOT_IMSG_TREE_ENTRIES, 0, 0, len);
+	if (wbuf == NULL)
+		return got_error_from_errno("imsg_create TREE_ENTRY");
+
+	ientries.nentries = idxN - idx0 + 1;
+	if (imsg_add(wbuf, &ientries, sizeof(ientries)) == -1)
+		return got_error_from_errno("imsg_add TREE_ENTRY");
+
+	for (i = idx0; i <= idxN; i++) {
+		struct got_parsed_tree_entry *pte = &entries[i];
+
+		/* Keep in sync with struct got_imsg_tree_object definition! */
+		if (imsg_add(wbuf, pte->id, SHA1_DIGEST_LENGTH) == -1)
+			return got_error_from_errno("imsg_add TREE_ENTRY");
+		if (imsg_add(wbuf, &pte->mode, sizeof(pte->mode)) == -1)
+			return got_error_from_errno("imsg_add TREE_ENTRY");
+		if (imsg_add(wbuf, &pte->namelen, sizeof(pte->namelen)) == -1)
+			return got_error_from_errno("imsg_add TREE_ENTRY");
+
+		/* Remaining bytes are the entry's name. */
+		if (imsg_add(wbuf, pte->name, pte->namelen) == -1)
+			return got_error_from_errno("imsg_add TREE_ENTRY");
+	}
+
+	wbuf->fd = -1;
+	imsg_close(ibuf, wbuf);
+	return NULL;
+}
+
+static const struct got_error *
+send_tree_entries(struct imsgbuf *ibuf, struct got_parsed_tree_entry *entries,
     int nentries)
 {
 	const struct got_error *err = NULL;
-	struct got_imsg_tree_object itree;
-	struct got_pathlist_entry *pe;
-	size_t totlen;
-	int nimsg; /* number of imsg queued in ibuf */
+	int i, j;
+	size_t entries_len = sizeof(struct got_imsg_tree_entries);
 
+	i = 0;
+	for (j = 0; j < nentries; j++) {
+		struct got_parsed_tree_entry *pte = &entries[j];
+		size_t len = sizeof(*pte) + pte->namelen;
+
+		if (j > 0 &&
+		    entries_len + len > MAX_IMSGSIZE - IMSG_HEADER_SIZE) {
+			err = send_tree_entries_batch(ibuf, entries,
+			    i, j - 1, entries_len);
+			if (err)
+				return err;
+			i = j;
+			entries_len = sizeof(struct got_imsg_tree_entries);
+		}
+
+		entries_len += len;
+	}
+
+	if (j > 0) {
+		err = send_tree_entries_batch(ibuf, entries, i, j - 1,
+		    entries_len);
+		if (err)
+			return err;
+	}
+
+	return NULL;
+}
+
+const struct got_error *
+got_privsep_send_tree(struct imsgbuf *ibuf,
+    struct got_parsed_tree_entry *entries, int nentries)
+{
+	const struct got_error *err = NULL;
+	struct got_imsg_tree_object itree;
+
+	memset(&itree, 0, sizeof(itree));
 	itree.nentries = nentries;
 	if (imsg_compose(ibuf, GOT_IMSG_TREE, 0, 0, -1, &itree, sizeof(itree))
 	    == -1)
 		return got_error_from_errno("imsg_compose TREE");
 
-	totlen = sizeof(itree);
-	nimsg = 1;
-	TAILQ_FOREACH(pe, entries, entry) {
-		const char *name = pe->path;
-		struct got_parsed_tree_entry *pte = pe->data;
-		struct ibuf *wbuf;
-		size_t namelen = strlen(name);
-		size_t len = sizeof(struct got_imsg_tree_entry) + namelen;
-
-		if (len > MAX_IMSGSIZE)
-			return got_error(GOT_ERR_NO_SPACE);
-
-		nimsg++;
-		if (totlen + len >= MAX_IMSGSIZE - (IMSG_HEADER_SIZE * nimsg)) {
-			err = flush_imsg(ibuf);
-			if (err)
-				return err;
-			nimsg = 0;
-		}
-
-		wbuf = imsg_create(ibuf, GOT_IMSG_TREE_ENTRY, 0, 0, len);
-		if (wbuf == NULL)
-			return got_error_from_errno("imsg_create TREE_ENTRY");
-
-		/* Keep in sync with struct got_imsg_tree_object definition! */
-		if (imsg_add(wbuf, pte->id, SHA1_DIGEST_LENGTH) == -1) {
-			err = got_error_from_errno("imsg_add TREE_ENTRY");
-			ibuf_free(wbuf);
-			return err;
-		}
-		if (imsg_add(wbuf, &pte->mode, sizeof(pte->mode)) == -1) {
-			err = got_error_from_errno("imsg_add TREE_ENTRY");
-			ibuf_free(wbuf);
-			return err;
-		}
-
-		if (imsg_add(wbuf, name, namelen) == -1) {
-			err = got_error_from_errno("imsg_add TREE_ENTRY");
-			ibuf_free(wbuf);
-			return err;
-		}
-
-		wbuf->fd = -1;
-		imsg_close(ibuf, wbuf);
-
-		totlen += len;
-	}
+	err = send_tree_entries(ibuf, entries, nentries);
+	if (err)
+		return err;
 
 	return flush_imsg(ibuf);
 }
 
+
+static const struct got_error *
+recv_tree_entries(void *data, size_t datalen, struct got_tree_object *tree,
+    int *nentries)
+{
+	const struct got_error *err = NULL;
+	struct got_imsg_tree_entries *ientries;
+	struct got_tree_entry *te;
+	size_t te_offset;
+	size_t i;
+
+	if (datalen <= sizeof(*ientries) ||
+	    datalen > MAX_IMSGSIZE - IMSG_HEADER_SIZE)
+		return got_error(GOT_ERR_PRIVSEP_LEN);
+
+	ientries = (struct got_imsg_tree_entries *)data;
+	if (ientries->nentries > INT_MAX) {
+		return got_error_msg(GOT_ERR_NO_SPACE,
+		    "too many tree entries");
+	}
+
+	te_offset = sizeof(*ientries);
+	for (i = 0; i < ientries->nentries; i++) {
+		struct got_imsg_tree_entry ite;
+		const char *te_name;
+		uint8_t *buf = (uint8_t *)data + te_offset;
+
+		if (te_offset >= datalen) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
+
+		/* Might not be aligned, size is ~32 bytes. */
+		memcpy(&ite, buf, sizeof(ite));
+
+		if (ite.namelen >= sizeof(te->name)) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
+		if (te_offset + sizeof(ite) + ite.namelen > datalen) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
+
+		if (*nentries >= tree->nentries) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
+		te = &tree->entries[*nentries];
+		te_name = buf + sizeof(ite);
+		memcpy(te->name, te_name, ite.namelen);
+		te->name[ite.namelen] = '\0';
+		memcpy(te->id.sha1, ite.id, SHA1_DIGEST_LENGTH);
+		te->mode = ite.mode;
+		te->idx = *nentries;
+		(*nentries)++;
+
+		te_offset += sizeof(ite) + ite.namelen;
+	}
+
+	return err;
+}
+
 const struct got_error *
 got_privsep_recv_tree(struct got_tree_object **tree, struct imsgbuf *ibuf)
 {
@@ -1565,7 +1587,7 @@ got_privsep_recv_tree(struct got_tree_object **tree, struct imsgbuf *ibuf)
 	int nentries = 0;
 
 	*tree = NULL;
-get_more:
+
 	err = read_imsg(ibuf);
 	if (err)
 		goto done;
@@ -1574,14 +1596,21 @@ get_more:
 		struct imsg imsg;
 		size_t n;
 		size_t datalen;
-		struct got_imsg_tree_entry *ite;
-		struct got_tree_entry *te = NULL;
 
 		n = imsg_get(ibuf, &imsg);
 		if (n == 0) {
-			if (*tree && (*tree)->nentries != nentries)
-				goto get_more;
-			break;
+			if ((*tree)) {
+				if (nentries < (*tree)->nentries) {
+					err = read_imsg(ibuf);
+					if (err)
+						break;
+					continue;
+				} else
+					break;
+			} else {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
 		}
 
 		if (imsg.hdr.len < IMSG_HEADER_SIZE + min_datalen) {
@@ -1607,6 +1636,10 @@ get_more:
 				break;
 			}
 			itree = imsg.data;
+			if (itree->nentries < 0) {
+				err = got_error(GOT_ERR_PRIVSEP_LEN);
+				break;
+			}
 			*tree = malloc(sizeof(**tree));
 			if (*tree == NULL) {
 				err = got_error_from_errno("malloc");
@@ -1616,42 +1649,21 @@ get_more:
 			    sizeof(struct got_tree_entry));
 			if ((*tree)->entries == NULL) {
 				err = got_error_from_errno("malloc");
+				free(*tree);
+				*tree = NULL;
 				break;
 			}
 			(*tree)->nentries = itree->nentries;
 			(*tree)->refcnt = 0;
 			break;
-		case GOT_IMSG_TREE_ENTRY:
+		case GOT_IMSG_TREE_ENTRIES:
 			/* This message should be preceeded by GOT_IMSG_TREE. */
 			if (*tree == NULL) {
 				err = got_error(GOT_ERR_PRIVSEP_MSG);
 				break;
 			}
-			if (datalen < sizeof(*ite) || datalen > MAX_IMSGSIZE) {
-				err = got_error(GOT_ERR_PRIVSEP_LEN);
-				break;
-			}
-
-			/* Remaining data contains the entry's name. */
-			datalen -= sizeof(*ite);
-			if (datalen == 0 || datalen > MAX_IMSGSIZE) {
-				err = got_error(GOT_ERR_PRIVSEP_LEN);
-				break;
-			}
-			ite = imsg.data;
-
-			if (datalen + 1 > sizeof(te->name)) {
-				err = got_error(GOT_ERR_NO_SPACE);
-				break;
-			}
-			te = &(*tree)->entries[nentries];
-			memcpy(te->name, imsg.data + sizeof(*ite), datalen);
-			te->name[datalen] = '\0';
-
-			memcpy(te->id.sha1, ite->id, SHA1_DIGEST_LENGTH);
-			te->mode = ite->mode;
-			te->idx = nentries;
-			nentries++;
+			err = recv_tree_entries(imsg.data, datalen,
+			    *tree, &nentries);
 			break;
 		default:
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -1679,6 +1691,7 @@ got_privsep_send_blob(struct imsgbuf *ibuf, size_t size, size_t hdrlen,
 {
 	struct got_imsg_blob iblob;
 
+	memset(&iblob, 0, sizeof(iblob));
 	iblob.size = size;
 	iblob.hdrlen = hdrlen;
 
@@ -1743,7 +1756,8 @@ got_privsep_recv_blob(uint8_t **outbuf, size_t *size, size_t *hdrlen,
 			break;
 		}
 
-		if (*size > GOT_PRIVSEP_INLINE_BLOB_DATA_MAX) {
+		if (*size > GOT_PRIVSEP_INLINE_BLOB_DATA_MAX ||
+		    *size > datalen + sizeof(*iblob)) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
 			break;
 		}
@@ -1884,21 +1898,10 @@ got_privsep_recv_tag(struct got_tag_object **tag, struct imsgbuf *ibuf)
 
 		memcpy((*tag)->id.sha1, itag->id, SHA1_DIGEST_LENGTH);
 
-		if (itag->tag_len == 0) {
-			(*tag)->tag = strdup("");
-			if ((*tag)->tag == NULL) {
-				err = got_error_from_errno("strdup");
-				break;
-			}
-		} else {
-			(*tag)->tag = malloc(itag->tag_len + 1);
-			if ((*tag)->tag == NULL) {
-				err = got_error_from_errno("malloc");
-				break;
-			}
-			memcpy((*tag)->tag, imsg.data + len,
-			    itag->tag_len);
-			(*tag)->tag[itag->tag_len] = '\0';
+		(*tag)->tag = strndup(imsg.data + len, itag->tag_len);
+		if ((*tag)->tag == NULL) {
+			err = got_error_from_errno("strndup");
+			break;
 		}
 		len += itag->tag_len;
 
@@ -1906,21 +1909,10 @@ got_privsep_recv_tag(struct got_tag_object **tag, struct imsgbuf *ibuf)
 		(*tag)->tagger_time = itag->tagger_time;
 		(*tag)->tagger_gmtoff = itag->tagger_gmtoff;
 
-		if (itag->tagger_len == 0) {
-			(*tag)->tagger = strdup("");
-			if ((*tag)->tagger == NULL) {
-				err = got_error_from_errno("strdup");
-				break;
-			}
-		} else {
-			(*tag)->tagger = malloc(itag->tagger_len + 1);
-			if ((*tag)->tagger == NULL) {
-				err = got_error_from_errno("malloc");
-				break;
-			}
-			memcpy((*tag)->tagger, imsg.data + len,
-			    itag->tagger_len);
-			(*tag)->tagger[itag->tagger_len] = '\0';
+		(*tag)->tagger = strndup(imsg.data + len, itag->tagger_len);
+		if ((*tag)->tagger == NULL) {
+			err = got_error_from_errno("strndup");
+			break;
 		}
 		len += itag->tagger_len;
 
@@ -1979,6 +1971,9 @@ got_privsep_init_pack_child(struct imsgbuf *ibuf, struct got_pack *pack,
 	struct got_imsg_pack ipack;
 	int fd;
 
+	memset(&ipackidx, 0, sizeof(ipackidx));
+	memset(&ipack, 0, sizeof(ipack));
+
 	ipackidx.len = packidx->len;
 	ipackidx.packfile_size = pack->filesize;
 	fd = dup(packidx->fd);
@@ -2017,6 +2012,7 @@ got_privsep_send_packed_obj_req(struct imsgbuf *ibuf, int idx,
 {
 	struct got_imsg_packed_object iobj;
 
+	memset(&iobj, 0, sizeof(iobj));
 	iobj.idx = idx;
 	memcpy(iobj.id, id->sha1, sizeof(iobj.id));
 
@@ -2034,6 +2030,7 @@ got_privsep_send_packed_raw_obj_req(struct imsgbuf *ibuf, int idx,
 {
 	struct got_imsg_packed_object iobj;
 
+	memset(&iobj, 0, sizeof(iobj));
 	iobj.idx = idx;
 	memcpy(iobj.id, id->sha1, sizeof(iobj.id));
 
@@ -2136,11 +2133,10 @@ got_privsep_recv_gitconfig_str(char **str, struct imsgbuf *ibuf)
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	const size_t min_datalen = 0;
 
 	*str = NULL;
 
-	err = got_privsep_recv_imsg(&imsg, ibuf, min_datalen);
+	err = got_privsep_recv_imsg(&imsg, ibuf, 0);
 	if (err)
 		return err;
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
@@ -2382,21 +2378,16 @@ got_privsep_recv_gotconfig_str(char **str, struct imsgbuf *ibuf)
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	const size_t min_datalen = 0;
 
 	*str = NULL;
 
-	err = got_privsep_recv_imsg(&imsg, ibuf, min_datalen);
+	err = got_privsep_recv_imsg(&imsg, ibuf, 0);
 	if (err)
 		return err;
 	datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 
 	switch (imsg.hdr.type) {
 	case GOT_IMSG_ERROR:
-		if (datalen < sizeof(struct got_imsg_error)) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
-			break;
-		}
 		err = recv_imsg_error(&imsg, datalen);
 		break;
 	case GOT_IMSG_GOTCONFIG_STR_VAL:
@@ -2443,10 +2434,6 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 
 	switch (imsg.hdr.type) {
 	case GOT_IMSG_ERROR:
-		if (datalen < sizeof(struct got_imsg_error)) {
-			err = got_error(GOT_ERR_PRIVSEP_LEN);
-			break;
-		}
 		err = recv_imsg_error(&imsg, datalen);
 		break;
 	case GOT_IMSG_GOTCONFIG_REMOTES:
@@ -2455,6 +2442,10 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 			break;
 		}
 		memcpy(&iremotes, imsg.data, sizeof(iremotes));
+		if (iremotes.nremotes < 0) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			break;
+		}
 		if (iremotes.nremotes == 0) {
 			imsg_free(&imsg);
 			return NULL;
@@ -2484,10 +2475,6 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 
 		switch (imsg.hdr.type) {
 		case GOT_IMSG_ERROR:
-			if (datalen < sizeof(struct got_imsg_error)) {
-				err = got_error(GOT_ERR_PRIVSEP_LEN);
-				break;
-			}
 			err = recv_imsg_error(&imsg, datalen);
 			break;
 		case GOT_IMSG_GOTCONFIG_REMOTE:
@@ -2621,7 +2608,6 @@ const struct got_error *
 got_privsep_send_commit_traversal_request(struct imsgbuf *ibuf,
     struct got_object_id *id, int idx, const char *path)
 {
-	const struct got_error *err = NULL;
 	struct ibuf *wbuf;
 	size_t path_len = strlen(path) + 1;
 
@@ -2630,21 +2616,15 @@ got_privsep_send_commit_traversal_request(struct imsgbuf *ibuf,
 	if (wbuf == NULL)
 		return got_error_from_errno(
 		    "imsg_create COMMIT_TRAVERSAL_REQUEST");
-	if (imsg_add(wbuf, id->sha1, SHA1_DIGEST_LENGTH) == -1) {
-		err = got_error_from_errno("imsg_add COMMIT_TRAVERSAL_REQUEST");
-		ibuf_free(wbuf);
-		return err;
-	}
-	if (imsg_add(wbuf, &idx, sizeof(idx)) == -1) {
-		err = got_error_from_errno("imsg_add COMMIT_TRAVERSAL_REQUEST");
-		ibuf_free(wbuf);
-		return err;
-	}
-	if (imsg_add(wbuf, path, path_len) == -1) {
-		err = got_error_from_errno("imsg_add COMMIT_TRAVERSAL_REQUEST");
-		ibuf_free(wbuf);
-		return err;
-	}
+	if (imsg_add(wbuf, id->sha1, SHA1_DIGEST_LENGTH) == -1)
+		return got_error_from_errno("imsg_add "
+		    "COMMIT_TRAVERSAL_REQUEST");
+	if (imsg_add(wbuf, &idx, sizeof(idx)) == -1)
+		return got_error_from_errno("imsg_add "
+		    "COMMIT_TRAVERSAL_REQUEST");
+	if (imsg_add(wbuf, path, path_len) == -1)
+		return got_error_from_errno("imsg_add "
+		    "COMMIT_TRAVERSAL_REQUEST");
 
 	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
@@ -2728,12 +2708,293 @@ got_privsep_recv_traversed_commits(struct got_commit_object **changed_commit,
 	return err;
 }
 
+const struct got_error *
+got_privsep_send_enumerated_tree(size_t *totlen, struct imsgbuf *ibuf,
+    struct got_object_id *tree_id, const char *path,
+    struct got_parsed_tree_entry *entries, int nentries)
+{
+	const struct got_error *err = NULL;
+	struct ibuf *wbuf;
+	size_t path_len = strlen(path);
+	size_t msglen;
+
+	msglen = sizeof(struct got_imsg_enumerated_tree) + path_len;
+	wbuf = imsg_create(ibuf, GOT_IMSG_ENUMERATED_TREE, 0, 0, msglen);
+	if (wbuf == NULL)
+		return got_error_from_errno("imsg_create ENUMERATED_TREE");
+
+	if (imsg_add(wbuf, tree_id->sha1, SHA1_DIGEST_LENGTH) == -1)
+		return got_error_from_errno("imsg_add ENUMERATED_TREE");
+	if (imsg_add(wbuf, &nentries, sizeof(nentries)) == -1)
+		return got_error_from_errno("imsg_add ENUMERATED_TREE");
+	if (imsg_add(wbuf, path, path_len) == -1)
+		return got_error_from_errno("imsg_add ENUMERATED_TREE");
+
+	wbuf->fd = -1;
+	imsg_close(ibuf, wbuf);
+
+	if (entries) {
+		err = send_tree_entries(ibuf, entries, nentries);
+		if (err)
+			return err;
+	}
+
+	return flush_imsg(ibuf);
+}
+
+const struct got_error *
+got_privsep_send_object_enumeration_request(struct imsgbuf *ibuf)
+{
+	if (imsg_compose(ibuf, GOT_IMSG_OBJECT_ENUMERATION_REQUEST,
+	    0, 0, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg_compose "
+		    "OBJECT_ENUMERATION_REQUEST");
+
+	return flush_imsg(ibuf);
+}
+
+const struct got_error *
+got_privsep_send_object_enumeration_done(struct imsgbuf *ibuf)
+{
+	if (imsg_compose(ibuf, GOT_IMSG_OBJECT_ENUMERATION_DONE,
+	    0, 0, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg_compose "
+		    "OBJECT_ENUMERATION_DONE");
+
+	return flush_imsg(ibuf);
+}
+
+const struct got_error *
+got_privsep_send_object_enumeration_incomplete(struct imsgbuf *ibuf)
+{
+	if (imsg_compose(ibuf, GOT_IMSG_OBJECT_ENUMERATION_INCOMPLETE,
+	    0, 0, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg_compose "
+		    "OBJECT_ENUMERATION_INCOMPLETE");
+
+	return flush_imsg(ibuf);
+}
+
+const struct got_error *
+got_privsep_send_enumerated_commit(struct imsgbuf *ibuf,
+    struct got_object_id *id, time_t mtime)
+{
+	struct ibuf *wbuf;
+
+	wbuf = imsg_create(ibuf, GOT_IMSG_ENUMERATED_COMMIT, 0, 0,
+	    sizeof(struct got_imsg_enumerated_commit) + SHA1_DIGEST_LENGTH);
+	if (wbuf == NULL)
+		return got_error_from_errno("imsg_create ENUMERATED_COMMIT");
+
+	/* Keep in sync with struct got_imsg_enumerated_commit! */
+	if (imsg_add(wbuf, id, SHA1_DIGEST_LENGTH) == -1)
+		return got_error_from_errno("imsg_add ENUMERATED_COMMIT");
+	if (imsg_add(wbuf, &mtime, sizeof(mtime)) == -1)
+		return got_error_from_errno("imsg_add ENUMERATED_COMMIT");
+
+	wbuf->fd = -1;
+	imsg_close(ibuf, wbuf);
+	/* Don't flush yet, tree entries or ENUMERATION_DONE will follow. */
+	return NULL;
+}
+
+const struct got_error *
+got_privsep_recv_enumerated_objects(int *found_all_objects,
+    struct imsgbuf *ibuf,
+    got_object_enumerate_commit_cb cb_commit,
+    got_object_enumerate_tree_cb cb_tree, void *cb_arg,
+    struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	struct imsg imsg;
+	struct got_imsg_enumerated_commit *icommit = NULL;
+	struct got_object_id commit_id;
+	int have_commit = 0;
+	time_t mtime = 0;
+	struct got_tree_object tree;
+	struct got_imsg_enumerated_tree *itree;
+	struct got_object_id tree_id;
+	char *path = NULL, *canon_path = NULL;
+	size_t datalen, path_len;
+	int nentries = -1;
+	int done = 0;
+
+	*found_all_objects = 0;
+	memset(&tree, 0, sizeof(tree));
+
+	while (!done) {
+		err = got_privsep_recv_imsg(&imsg, ibuf, 0);
+		if (err)
+			break;
+
+		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
+		switch (imsg.hdr.type) {
+		case GOT_IMSG_ENUMERATED_COMMIT:
+			if (have_commit && nentries != -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			if (datalen != sizeof(*icommit)) {
+				err = got_error(GOT_ERR_PRIVSEP_LEN);
+				break;
+			}
+			icommit = (struct got_imsg_enumerated_commit *)imsg.data;
+			memcpy(commit_id.sha1, icommit->id, SHA1_DIGEST_LENGTH);
+			mtime = icommit->mtime;
+			have_commit = 1;
+			break;
+		case GOT_IMSG_ENUMERATED_TREE:
+			/* Should be preceeded by GOT_IMSG_ENUMERATED_COMMIT. */
+			if (!have_commit) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			if (datalen < sizeof(*itree)) {
+				err = got_error(GOT_ERR_PRIVSEP_LEN);
+				break;
+			}
+			itree = imsg.data;
+			path_len = datalen - sizeof(*itree);
+			if (path_len == 0) {
+				err = got_error(GOT_ERR_PRIVSEP_LEN);
+				break;
+			}
+			memcpy(tree_id.sha1, itree->id, sizeof(tree_id.sha1));
+			free(path);
+			path = malloc(path_len + 1);
+			if (path == NULL) {
+				err = got_error_from_errno("malloc");
+				break;
+			}
+			free(canon_path);
+			canon_path = malloc(path_len + 1);
+			if (canon_path == NULL) {
+				err = got_error_from_errno("malloc");
+				break;
+			}
+			memcpy(path, (uint8_t *)imsg.data + sizeof(*itree),
+			    path_len);
+			path[path_len] = '\0';
+			if (!got_path_is_absolute(path)) {
+				err = got_error(GOT_ERR_BAD_PATH);
+				break;
+			}
+			if (got_path_is_root_dir(path)) {
+				/* XXX check what got_canonpath() does wrong */
+				canon_path[0] = '/';
+				canon_path[1] = '\0';
+			} else {
+				err = got_canonpath(path, canon_path,
+				    path_len + 1);
+				if (err)
+					break;
+			}
+			if (strcmp(path, canon_path) != 0) {
+				err = got_error(GOT_ERR_BAD_PATH);
+				break;
+			}
+			if (nentries != -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			if (itree->nentries < -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			if (itree->nentries == -1) {
+				/* Tree was not found in pack file. */
+				err = cb_tree(cb_arg, NULL, mtime, &tree_id,
+				    path, repo);
+				break;
+			}
+			if (itree->nentries > INT_MAX) {
+				err = got_error(GOT_ERR_PRIVSEP_LEN);
+				break;
+			}
+			tree.entries = calloc(itree->nentries,
+			    sizeof(struct got_tree_entry));
+			if (tree.entries == NULL) {
+				err = got_error_from_errno("calloc");
+				break;
+			}
+			if (itree->nentries == 0) {
+				err = cb_tree(cb_arg, &tree, mtime, &tree_id,
+				    path, repo);
+				if (err)
+					break;
+
+				/* Prepare for next tree. */
+				free(tree.entries);
+				memset(&tree, 0, sizeof(tree));
+				nentries = -1;
+			} else {
+				tree.nentries = itree->nentries;
+				nentries = 0;
+			}
+			break;
+		case GOT_IMSG_TREE_ENTRIES:
+			/* Should be preceeded by GOT_IMSG_ENUMERATED_TREE. */
+			if (nentries <= -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = recv_tree_entries(imsg.data, datalen,
+			    &tree, &nentries);
+			if (err)
+				break;
+			if (tree.nentries == nentries) {
+				err = cb_tree(cb_arg, &tree, mtime, &tree_id,
+				    path, repo);
+				if (err)
+					break;
+
+				/* Prepare for next tree. */
+				free(tree.entries);
+				memset(&tree, 0, sizeof(tree));
+				nentries = -1;
+			}
+			break;
+		case GOT_IMSG_TREE_ENUMERATION_DONE:
+			/* All trees have been found and traversed. */
+			if (!have_commit || path == NULL || nentries != -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = cb_commit(cb_arg, mtime, &commit_id, repo);
+			if (err)
+				break;
+			have_commit = 0;
+			break;
+		case GOT_IMSG_OBJECT_ENUMERATION_DONE:
+			*found_all_objects = 1;
+			done = 1;
+			break;
+		case GOT_IMSG_OBJECT_ENUMERATION_INCOMPLETE:
+			done = 1;
+			break;
+		default:
+			err = got_error(GOT_ERR_PRIVSEP_MSG);
+			break;
+		}
+
+		imsg_free(&imsg);
+		if (err)
+			break;
+	}
+
+	free(path);
+	free(canon_path);
+	free(tree.entries);
+	return err;
+}
+
 const struct got_error *
 got_privsep_send_raw_delta_req(struct imsgbuf *ibuf, int idx,
     struct got_object_id *id)
 {
 	struct got_imsg_raw_delta_request dreq;
 
+	memset(&dreq, 0, sizeof(dreq));
 	dreq.idx = idx;
 	memcpy(dreq.id, id->sha1, SHA1_DIGEST_LENGTH);
 
@@ -2758,6 +3019,7 @@ got_privsep_send_raw_delta(struct imsgbuf *ibuf, uint64_t base_size,
 	struct got_imsg_raw_delta idelta;
 	int ret;
 
+	memset(&idelta, 0, sizeof(idelta));
 	idelta.base_size = base_size;
 	idelta.result_size = result_size;
 	idelta.delta_size = delta_size;
@@ -2832,15 +3094,16 @@ got_privsep_recv_raw_delta(uint64_t *base_size, uint64_t *result_size,
 	return err;
 }
 
-const struct got_error *
-got_privsep_send_object_idlist(struct imsgbuf *ibuf,
-    struct got_object_id **ids, size_t nids)
+static const struct got_error *
+send_idlist(struct imsgbuf *ibuf, struct got_object_id **ids, size_t nids)
 {
 	const struct got_error *err = NULL;
 	struct got_imsg_object_idlist idlist;
 	struct ibuf *wbuf;
 	size_t i;
 
+	memset(&idlist, 0, sizeof(idlist));
+
 	if (nids > GOT_IMSG_OBJ_ID_LIST_MAX_NIDS)
 		return got_error(GOT_ERR_NO_SPACE);
 
@@ -2852,27 +3115,49 @@ got_privsep_send_object_idlist(struct imsgbuf *ibuf,
 	}
 
 	idlist.nids = nids;
-	if (imsg_add(wbuf, &idlist, sizeof(idlist)) == -1) {
-		err = got_error_from_errno("imsg_add OBJ_ID_LIST");
-		ibuf_free(wbuf);
-		return err;
-	}
+	if (imsg_add(wbuf, &idlist, sizeof(idlist)) == -1)
+		return got_error_from_errno("imsg_add OBJ_ID_LIST");
 
 	for (i = 0; i < nids; i++) {
 		struct got_object_id *id = ids[i];
-		if (imsg_add(wbuf, id, sizeof(*id)) == -1) {
-			err = got_error_from_errno("imsg_add OBJ_ID_LIST");
-			ibuf_free(wbuf);
-			return err;
-		}
+		if (imsg_add(wbuf, id, sizeof(*id)) == -1)
+			return got_error_from_errno("imsg_add OBJ_ID_LIST");
 	}
-	
+
 	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	return flush_imsg(ibuf);
 }
 
+const struct got_error *
+got_privsep_send_object_idlist(struct imsgbuf *ibuf,
+    struct got_object_id **ids, size_t nids)
+{
+	const struct got_error *err = NULL;
+	struct got_object_id *idlist[GOT_IMSG_OBJ_ID_LIST_MAX_NIDS];
+	int i, queued = 0;
+
+	for (i = 0; i < nids; i++) {
+		idlist[i % nitems(idlist)] = ids[i];
+		queued++;
+		if (queued >= nitems(idlist)) {
+			err = send_idlist(ibuf, idlist, queued);
+			if (err)
+				return err;
+			queued = 0;
+		}
+	}
+
+	if (queued > 0) {
+		err = send_idlist(ibuf, idlist, queued);
+		if (err)
+			return err;
+	}
+
+	return NULL;
+}
+
 const struct got_error *
 got_privsep_send_object_idlist_done(struct imsgbuf *ibuf)
 {
@@ -2908,7 +3193,8 @@ got_privsep_recv_object_idlist(int *done, struct got_object_id **ids,
 			break;
 		}
 		idlist = imsg.data;
-		if (idlist->nids > GOT_IMSG_OBJ_ID_LIST_MAX_NIDS) {
+		if (idlist->nids > GOT_IMSG_OBJ_ID_LIST_MAX_NIDS ||
+		    idlist->nids * sizeof(**ids) > datalen - sizeof(*idlist)) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
 			break;
 		}
@@ -2918,7 +3204,7 @@ got_privsep_recv_object_idlist(int *done, struct got_object_id **ids,
 			err = got_error_from_errno("calloc");
 			break;
 		}
-		memcpy(*ids, (uint8_t *)imsg.data + sizeof(idlist),
+		memcpy(*ids, (uint8_t *)imsg.data + sizeof(*idlist),
 		    *nids * sizeof(**ids));
 		break;
 	case GOT_IMSG_OBJ_ID_LIST_DONE:
@@ -2953,6 +3239,8 @@ got_privsep_send_reused_deltas(struct imsgbuf *ibuf,
 	struct got_imsg_reused_deltas ideltas;
 	size_t i;
 
+	memset(&ideltas, 0, sizeof(ideltas));
+
 	if (ndeltas > GOT_IMSG_REUSED_DELTAS_MAX_NDELTAS)
 		return got_error(GOT_ERR_NO_SPACE);
 
@@ -2964,21 +3252,15 @@ got_privsep_send_reused_deltas(struct imsgbuf *ibuf,
 	}
 
 	ideltas.ndeltas = ndeltas;
-	if (imsg_add(wbuf, &ideltas, sizeof(ideltas)) == -1) {
-		err = got_error_from_errno("imsg_add REUSED_DELTAS");
-		ibuf_free(wbuf);
-		return err;
-	}
+	if (imsg_add(wbuf, &ideltas, sizeof(ideltas)) == -1)
+		return got_error_from_errno("imsg_add REUSED_DELTAS");
 
 	for (i = 0; i < ndeltas; i++) {
 		struct got_imsg_reused_delta *delta = &deltas[i];
-		if (imsg_add(wbuf, delta, sizeof(*delta)) == -1) {
-			err = got_error_from_errno("imsg_add REUSED_DELTAS");
-			ibuf_free(wbuf);
-			return err;
-		}
+		if (imsg_add(wbuf, delta, sizeof(*delta)) == -1)
+			return got_error_from_errno("imsg_add REUSED_DELTAS");
 	}
-	
+
 	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
@@ -3019,12 +3301,14 @@ got_privsep_recv_reused_deltas(int *done, struct got_imsg_reused_delta *deltas,
 			break;
 		}
 		ideltas = imsg.data;
-		if (ideltas->ndeltas > GOT_IMSG_OBJ_ID_LIST_MAX_NIDS) {
+		if (ideltas->ndeltas > GOT_IMSG_OBJ_ID_LIST_MAX_NIDS ||
+		    ideltas->ndeltas * sizeof(*deltas) >
+		    datalen - sizeof(*ideltas)) {
 			err = got_error(GOT_ERR_PRIVSEP_LEN);
 			break;
 		}
 		*ndeltas = ideltas->ndeltas;
-		memcpy(deltas, (uint8_t *)imsg.data + sizeof(ideltas),
+		memcpy(deltas, (uint8_t *)imsg.data + sizeof(*ideltas),
 		    *ndeltas * sizeof(*deltas));
 		break;
 	case GOT_IMSG_DELTA_REUSE_DONE:
