@@ -44,6 +44,7 @@
 #include "got_path.h"
 #include "got_cancel.h"
 #include "got_object.h"
+#include "got_opentemp.h"
 
 #include "got_lib_delta.h"
 #include "got_lib_inflate.h"
@@ -61,6 +62,8 @@
 #define nitems(_a) (sizeof(_a) / sizeof((_a)[0]))
 #endif
 
+#define GOT_PACK_NUM_TEMPFILES		GOT_PACK_CACHE_SIZE * 2
+
 RB_PROTOTYPE(got_packidx_bloom_filter_tree, got_packidx_bloom_filter, entry,
     got_packidx_bloom_filter_cmp);
 
@@ -239,6 +242,46 @@ done:
 
 }
 
+const struct got_error *
+got_repo_pack_fds_open(int **pack_fds)
+{
+	const struct got_error *err = NULL;
+	int i, pack_fds_tmp[GOT_PACK_NUM_TEMPFILES];
+
+	*pack_fds = calloc(GOT_PACK_NUM_TEMPFILES, sizeof(**pack_fds));
+	if (*pack_fds == NULL)
+		return got_error_from_errno("calloc");
+
+	for (i = 0; i < GOT_PACK_NUM_TEMPFILES; i++) {
+		pack_fds_tmp[i] = got_opentempfd();
+		if (pack_fds_tmp[i] == -1) {
+			err = got_error_from_errno("got_opentempfd");
+			got_repo_pack_fds_close(pack_fds_tmp);
+			return err;
+		}
+	}
+	memcpy(*pack_fds, pack_fds_tmp, sizeof(pack_fds_tmp));
+	return err;
+}
+
+const struct got_error *
+got_repo_pack_fds_close(int *pack_fds)
+{
+	const struct got_error *err = NULL;
+	int i;
+
+	for (i = 0; i < GOT_PACK_NUM_TEMPFILES; i++) {
+		if (pack_fds[i] == -1)
+			continue;
+		if (close(pack_fds[i]) == -1) {
+			err = got_error_from_errno("close");
+			break;
+		}
+	}
+	free(pack_fds);
+	return err;
+}
+
 const struct got_error *
 got_repo_cache_object(struct got_repository *repo, struct got_object_id *id,
     struct got_object *obj)
@@ -644,12 +687,12 @@ static const char *const repo_extensions[] = {
 
 const struct got_error *
 got_repo_open(struct got_repository **repop, const char *path,
-    const char *global_gitconfig_path)
+    const char *global_gitconfig_path, int *pack_fds)
 {
 	struct got_repository *repo = NULL;
 	const struct got_error *err = NULL;
 	char *repo_path = NULL;
-	size_t i;
+	size_t i, j = 0;
 	struct rlimit rl;
 
 	*repop = NULL;
@@ -694,6 +737,15 @@ got_repo_open(struct got_repository **repop, const char *path,
 	repo->pack_cache_size = GOT_PACK_CACHE_SIZE;
 	if (repo->pack_cache_size > rl.rlim_cur / 8)
 		repo->pack_cache_size = rl.rlim_cur / 8;
+	for (i = 0; i < nitems(repo->packs); i++) {
+		if (i < repo->pack_cache_size) {
+			repo->packs[i].basefd = pack_fds[j++];
+			repo->packs[i].accumfd = pack_fds[j++];
+		} else {
+			repo->packs[i].basefd = -1;
+			repo->packs[i].accumfd = -1;
+		}
+	}
 
 	repo_path = realpath(path, NULL);
 	if (repo_path == NULL) {
@@ -776,11 +828,10 @@ got_repo_close(struct got_repository *repo)
 		free(bf);
 	}
 
-	for (i = 0; i < repo->pack_cache_size; i++) {
-		if (repo->packs[i].path_packfile == NULL)
-			break;
-		got_pack_close(&repo->packs[i]);
-	}
+	for (i = 0; i < repo->pack_cache_size; i++)
+		if (repo->packs[i].path_packfile)
+			if (repo->packs[i].path_packfile)
+				got_pack_close(&repo->packs[i]);
 
 	free(repo->path);
 	free(repo->path_git_dir);
@@ -1318,11 +1369,18 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 	}
 
 	if (i == repo->pack_cache_size) {
+		struct got_pack tmp;
 		err = got_pack_close(&repo->packs[i - 1]);
 		if (err)
 			return err;
-		memmove(&repo->packs[1], &repo->packs[0],
-		    sizeof(repo->packs) - sizeof(repo->packs[0]));
+		if (ftruncate(repo->packs[i - 1].basefd, 0L) == -1)
+			return got_error_from_errno("ftruncate");
+		if (ftruncate(repo->packs[i - 1].accumfd, 0L) == -1)
+			return got_error_from_errno("ftruncate");
+		memcpy(&tmp, &repo->packs[i - 1], sizeof(tmp)); 
+		memcpy(&repo->packs[i - 1], &repo->packs[0],
+		    sizeof(repo->packs[i - 1]));
+		memcpy(&repo->packs[0], &tmp, sizeof(repo->packs[0]));
 		i = 0;
 	}
 
