@@ -55,11 +55,13 @@
 #include "got_lib_delta.h"
 #include "got_lib_object.h"
 #include "got_lib_privsep.h"
+#include "got_lib_sha1.h"
 
 struct imsgbuf ibuf;
 
 static const struct got_error *
-send_patch(const char *oldname, const char *newname, int git)
+send_patch(const char *oldname, const char *newname, const char *commitid,
+    const char *blob, int git)
 {
 	struct got_imsg_patch p;
 
@@ -71,9 +73,13 @@ send_patch(const char *oldname, const char *newname, int git)
 	if (newname != NULL)
 		strlcpy(p.new, newname, sizeof(p.new));
 
+	if (commitid != NULL && blob != NULL) {
+		strlcpy(p.cid, commitid, sizeof(p.cid));
+		strlcpy(p.blob, blob, sizeof(p.blob));
+	}
+
 	p.git = git;
-	if (imsg_compose(&ibuf, GOT_IMSG_PATCH, 0, 0, -1,
-	    &p, sizeof(p)) == -1)
+	if (imsg_compose(&ibuf, GOT_IMSG_PATCH, 0, 0, -1, &p, sizeof(p)) == -1)
 		return got_error_from_errno("imsg_compose GOT_IMSG_PATCH");
 	return NULL;
 }
@@ -121,11 +127,32 @@ filename(const char *at, char **name)
 	return NULL;
 }
 
+static const struct got_error *
+blobid(const char *line, char **blob)
+{
+	uint8_t digest[SHA1_DIGEST_LENGTH];
+	size_t len;
+
+	*blob = NULL;
+
+	len = strspn(line, "0123456789abcdefABCDEF");
+	if ((*blob = strndup(line, len)) == NULL)
+		return got_error_from_errno("strndup");
+
+	if (!got_parse_sha1_digest(digest, *blob)) {
+		/* silently ignore invalid blob ids */
+		free(*blob);
+		*blob = NULL;
+	}
+	return NULL;
+}
+
 static const struct got_error *
 find_patch(int *done, FILE *fp)
 {
 	const struct got_error *err = NULL;
 	char	*old = NULL, *new = NULL;
+	char	*commitid = NULL, *blob = NULL;
 	char	*line = NULL;
 	size_t	 linesize = 0;
 	ssize_t	 linelen;
@@ -146,13 +173,24 @@ find_patch(int *done, FILE *fp)
 		} else if (!strncmp(line, "+++ ", 4)) {
 			free(new);
 			err = filename(line+4, &new);
+		} else if (!git && !strncmp(line, "blob - ", 7)) {
+			free(blob);
+			err = blobid(line + 7, &blob);
 		} else if (rename && !strncmp(line, "rename to ", 10)) {
 			free(new);
 			err = filename(line + 10, &new);
 		} else if (git && !strncmp(line, "similarity index 100%", 21))
 			rename = 1;
-		else if (!strncmp(line, "diff --git a/", 13))
+		else if (!strncmp(line, "diff --git a/", 13)) {
 			git = 1;
+			free(commitid);
+			commitid = NULL;
+			free(blob);
+			blob = NULL;
+		} else if (!git && !strncmp(line, "diff ", 5)) {
+			free(commitid);
+			err = blobid(line + 5, &commitid);
+		}
 
 		if (err)
 			break;
@@ -164,7 +202,8 @@ find_patch(int *done, FILE *fp)
 		 */
 		if (rename && old != NULL && new != NULL) {
 			*done = 1;
-			err = send_patch(old, new, git);
+			err = send_patch(old, new, commitid,
+			    blob, git);
 			break;
 		}
 
@@ -174,20 +213,23 @@ find_patch(int *done, FILE *fp)
 			    (!create && old == NULL))
 				err = got_error(GOT_ERR_PATCH_MALFORMED);
 			else
-				err = send_patch(old, new, git);
+				err = send_patch(old, new, commitid,
+				    blob, git);
 
 			if (err)
 				break;
 
 			/* rewind to previous line */
-			if (fseek(fp, linelen * -1, SEEK_CUR) == -1)
-				err = got_error_from_errno("fseek");
+			if (fseeko(fp, -linelen, SEEK_CUR) == -1)
+				err = got_error_from_errno("fseeko");
 			break;
 		}
 	}
 
 	free(old);
 	free(new);
+	free(commitid);
+	free(blob);
 	free(line);
 	if (ferror(fp) && err == NULL)
 		err = got_error_from_errno("getline");
@@ -197,7 +239,7 @@ find_patch(int *done, FILE *fp)
 }
 
 static const struct got_error *
-strtolnum(char **str, long *n)
+strtolnum(char **str, int *n)
 {
 	char		*p, c;
 	const char	*errstr;
@@ -208,7 +250,7 @@ strtolnum(char **str, long *n)
 	c = *p;
 	*p = '\0';
 
-	*n = strtonum(*str, 0, LONG_MAX, &errstr);
+	*n = strtonum(*str, 0, INT_MAX, &errstr);
 	if (errstr != NULL)
 		return got_error(GOT_ERR_PATCH_MALFORMED);
 
@@ -263,10 +305,11 @@ parse_hdr(char *s, int *done, struct got_imsg_patch_hunk *hdr)
 	if (*s != '@')
 		return got_error(GOT_ERR_PATCH_MALFORMED);
 
-	if (hdr->oldfrom >= LONG_MAX - hdr->oldlines ||
-	    hdr->newfrom >= LONG_MAX - hdr->newlines ||
+	if (hdr->oldfrom >= INT_MAX - hdr->oldlines ||
+	    hdr->newfrom >= INT_MAX - hdr->newlines ||
 	    /* not so sure about this one */
-	    hdr->oldlines >= LONG_MAX - hdr->newlines - 1)
+	    hdr->oldlines >= INT_MAX - hdr->newlines - 1 ||
+	    (hdr->oldlines == 0 && hdr->newlines == 0))
 		return got_error(GOT_ERR_PATCH_MALFORMED);
 
 	if (hdr->oldlines == 0) {
@@ -338,7 +381,7 @@ parse_hunk(FILE *fp, int *done)
 	char	*line = NULL, ch;
 	size_t	 linesize = 0;
 	ssize_t	 linelen;
-	long	 leftold, leftnew;
+	int	 leftold, leftnew;
 
 	linelen = getline(&line, &linesize, fp);
 	if (linelen == -1) {
@@ -350,8 +393,8 @@ parse_hunk(FILE *fp, int *done)
 	if (err)
 		goto done;
 	if (*done) {
-		if (fseek(fp, linelen * -1, SEEK_CUR) == -1)
-			err = got_error_from_errno("fseek");
+		if (fseeko(fp, -linelen, SEEK_CUR) == -1)
+			err = got_error_from_errno("fseeko");
 		goto done;
 	}
 
